import unittest

import numpy as np
import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class RandomFlipTest(tf.test.TestCase, parameterized.TestCase):
    def test_horizontal_flip(self):
        np.random.seed(1337)
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        inp = np.random.random((2, 5, 8, 3))
        expected_output = np.flip(inp, axis=2)
        layer = layers.RandomFlip("horizontal")
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(inp)
            self.assertAllClose(expected_output, actual_output)

    def test_flip_ragged(self):
        images = tf.ragged.stack([tf.ones((4, 4, 3)), tf.ones((4, 8, 3))])
        bounding_boxes = {
            "boxes": tf.ragged.stack([tf.ones((5, 4)), tf.ones((3, 4))]),
            "classes": tf.ragged.stack([tf.ones((5,)), tf.ones((3,))]),
        }
        inputs = {"images": images, "bounding_boxes": bounding_boxes}
        layer = layers.RandomFlip(mode="horizontal", bounding_box_format="xywh")
        _ = layer(inputs)

    def test_vertical_flip(self):
        np.random.seed(1337)
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        inp = np.random.random((2, 5, 8, 3))
        expected_output = np.flip(inp, axis=1)
        layer = layers.RandomFlip("vertical")
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(inp)
            self.assertAllClose(expected_output, actual_output)

    def test_flip_both(self):
        np.random.seed(1337)
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        inp = np.random.random((2, 5, 8, 3))
        expected_output = np.flip(inp, axis=2)
        expected_output = np.flip(expected_output, axis=1)
        layer = layers.RandomFlip("horizontal_and_vertical")
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(inp)
        self.assertAllClose(expected_output, actual_output)

    def test_random_flip_default(self):
        input_images = np.random.random((2, 5, 8, 3)).astype(np.float32)
        expected_output = np.flip(input_images, axis=2)
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        layer = layers.RandomFlip()
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(input_images)
            self.assertAllClose(expected_output, actual_output)

    def test_random_flip_low_rate(self):
        input_images = np.random.random((2, 5, 8, 3)).astype(np.float32)
        expected_output = input_images
        # mock_random > 0.5 but no flipping occurs due to low rate
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        layer = layers.RandomFlip(rate=0.1)
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(input_images)
        self.assertAllClose(expected_output, actual_output)

    def test_random_flip_high_rate(self):
        input_images = np.random.random((2, 5, 8, 3)).astype(np.float32)
        expected_output = np.flip(input_images, axis=2)
        # mock_random is small (0.2) but flipping still occurs due to high rate
        mock_random = tf.convert_to_tensor([[0.2], [0.2]])
        layer = layers.RandomFlip(rate=0.9)
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(input_images)
        self.assertAllClose(expected_output, actual_output)

    def test_config_with_custom_name(self):
        layer = layers.RandomFlip(name="image_preproc")
        config = layer.get_config()
        layer_1 = layers.RandomFlip.from_config(config)
        self.assertEqual(layer_1.name, layer.name)

    def test_random_flip_unbatched_image(self):
        input_image = np.random.random((4, 4, 1)).astype(np.float32)
        expected_output = np.flip(input_image, axis=0)
        mock_random = tf.convert_to_tensor([[0.6]])
        layer = layers.RandomFlip("vertical")
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            actual_output = layer(input_image)
            self.assertAllClose(expected_output, actual_output)

    def test_output_dtypes(self):
        inputs = np.array([[[1], [2]], [[3], [4]]], dtype="float64")
        layer = layers.RandomFlip()
        self.assertAllEqual(layer(inputs).dtype, "float32")
        layer = layers.RandomFlip(dtype="uint8")
        self.assertAllEqual(layer(inputs).dtype, "uint8")

    def test_augment_bounding_box_batched_input(self):
        image = tf.zeros([20, 20, 3])
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 10, 10], [4, 4, 12, 12]],
                    [[4, 4, 12, 12], [0, 0, 10, 10]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.convert_to_tensor(
                [
                    [0, 0],
                    [0, 0],
                ]
            ),
        }

        input = {"images": [image, image], "bounding_boxes": bounding_boxes}
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        layer = layers.RandomFlip(
            "horizontal_and_vertical", bounding_box_format="xyxy"
        )
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            output = layer(input)

        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[10, 10, 20, 20], [8, 8, 16, 16]],
                    [[8, 8, 16, 16], [10, 10, 20, 20]],
                ]
            ),
            "classes": tf.convert_to_tensor(
                [
                    [0, 0],
                    [0, 0],
                ]
            ),
        }
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )
        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 10, 10], [4, 4, 12, 12]], [[0, 0, 10, 10]]],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 0], [0]], dtype=tf.float32),
        }

        input = {"images": image, "bounding_boxes": bounding_boxes}
        mock_random = tf.convert_to_tensor([[0.6], [0.6]])
        layer = layers.RandomFlip(
            "horizontal_and_vertical", bounding_box_format="xyxy"
        )
        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            output = layer(input)

        expected_output = {
            "boxes": tf.ragged.constant(
                [[[10, 10, 20, 20], [8, 8, 16, 16]], [[10, 10, 20, 20]]],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 0], [0]], dtype=tf.float32),
        }

        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )
        expected_output = bounding_box.to_dense(expected_output)
        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_segmentation_mask(self):
        np.random.seed(1337)
        image = np.random.random((1, 20, 20, 3)).astype(np.float32)
        mask = np.random.randint(2, size=(1, 20, 20, 1)).astype(np.float32)

        input = {"images": image, "segmentation_masks": mask}

        # Flip both vertically and horizontally
        mock_random = tf.convert_to_tensor([[0.6]])
        layer = layers.RandomFlip("horizontal_and_vertical")

        with unittest.mock.patch.object(
            layer._random_generator,
            "random_uniform",
            return_value=mock_random,
        ):
            output = layer(input)

        expected_mask = np.flip(np.flip(mask, axis=1), axis=2)

        self.assertAllClose(expected_mask, output["segmentation_masks"])

    def test_ragged_bounding_boxes(self):
        input_image = np.random.random((2, 512, 512, 3)).astype(np.float32)
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[200, 200, 400, 400], [100, 100, 300, 300]],
                    [[200, 200, 400, 400]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 0], [0]], dtype=tf.float32),
        }

        input = {"images": input_image, "bounding_boxes": bounding_boxes}
        layer = layers.RandomFlip(bounding_box_format="xyxy")
        _ = layer(input)

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        layer = layers.RandomFlip()

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        layer = layers.RandomFlip()

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(
            isinstance(result["segmentation_masks"], tf.RaggedTensor)
        )
        self.assertAllInSet(
            result["segmentation_masks"].to_tensor(0), tf.range(0, 10)
        )



import tensorflow as tf

from keras_aug import layers


class RescaleTest(tf.test.TestCase):
    regular_args = {
        "scale": 1.0 / 255.0,
        "offset": 0.0,
    }
    no_aug_args = {
        "scale": 1.0,
        "offset": 0.0,
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape)

        layer = layers.Rescale(**self.no_aug_args)
        output = layer(image)

        self.assertAllClose(image, output)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.Rescale(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_rescale_output(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        scale = tf.convert_to_tensor(self.regular_args["scale"])
        scale = tf.reshape(scale, shape=(1, 1, 1, 1))
        offset = tf.convert_to_tensor(self.regular_args["offset"])
        offset = tf.reshape(offset, shape=(1, 1, 1, 1))
        expected_output = image * scale + offset

        layer = layers.Rescale(**self.regular_args)
        output = layer(image)

        self.assertAllClose(output, expected_output)



import tensorflow as tf

from keras_aug import layers


class ChannelShuffleTest(tf.test.TestCase):
    def test_channel_shuffle_call_results_one_channel(self):
        xs = tf.cast(
            tf.stack(
                [3 * tf.ones((40, 40, 1)), 2 * tf.ones((40, 40, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )

        layer = layers.ChannelShuffle(groups=1)
        xs = layer(xs)
        self.assertTrue(tf.math.reduce_any(xs[0] == 3.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))

    def test_channel_shuffle_call_results_multi_channel(self):
        xs = tf.cast(
            tf.stack(
                [3 * tf.ones((40, 40, 20)), 2 * tf.ones((40, 40, 20))],
                axis=0,
            ),
            dtype=tf.float32,
        )

        layer = layers.ChannelShuffle(groups=5)
        xs = layer(xs)
        self.assertTrue(tf.math.reduce_any(xs[0] == 3.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))

    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((128, 64, 1)), tf.ones((128, 64, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )

        layer = layers.ChannelShuffle(groups=1)
        xs = layer(xs)
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_single_image(self):
        xs = tf.cast(tf.ones((128, 128, 1)), dtype=tf.float32)

        layer = layers.ChannelShuffle(groups=1)
        xs = layer(xs)
        self.assertTrue(tf.math.reduce_any(xs == 1.0))



import math

import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomRotate(VectorizedBaseRandomLayer):
    """Randomly rotates the input images.

    The unit of the factor is degree. A positive value means rotating counter
    clock-wise, while a negative value means clock-wise.

    Args:
        factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range of the degree for random rotation. When represented as a
            single float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. A positive value means rotating
            counter clock-wise, while a negative value means clock-wise.
        interpolation (str, optional): The interpolation mode. Supported values:
            ``"nearest", "bilinear"``. Defaults to `"bilinear"`.
        fill_mode (str, optional): The fill mode. Supported values:
            ``"constant", "reflect", "wrap", "nearest"``. Defaults to
            ``"constant"``.
        fill_value (int|float, optional): The value to be filled outside the
            boundaries when ``fill_mode="constant"``. Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        factor,
        interpolation="bilinear",
        fill_mode="constant",
        fill_value=0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.factor = augmentation_utils.parse_factor(
            factor,
            min_value=-180,
            max_value=180,
            center_value=0,
            seed=seed,
        )

        augmentation_utils.check_fill_mode_and_interpolation(
            fill_mode, interpolation
        )
        self.interpolation = interpolation
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.bounding_box_format = bounding_box_format
        self.seed = seed

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # cast to float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        angles = self.factor(shape=(batch_size, 1), dtype=tf.float32)
        angles = angles / 360.0 * 2.0 * math.pi
        rotation_matrixes = image_utils.get_rotation_matrix(
            angles, heights, widths, to_square=True
        )
        # (batch_size, 3, 3)
        return {
            "angles": angles,
            "rotation_matrixes": rotation_matrixes,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        original_shape = images.shape
        batch_size = tf.shape(images)[0]
        rotation_matrixes = transformations["rotation_matrixes"]
        rotation_matrixes = tf.reshape(
            rotation_matrixes, shape=(batch_size, -1)
        )
        rotation_matrixes = rotation_matrixes[:, :-1]

        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if images.dtype == tf.bfloat16:
            images = tf.cast(images, dtype=tf.float32)
        images = image_utils.transform(
            images,
            rotation_matrixes,
            fill_mode=self.fill_mode,
            fill_value=self.fill_value,
            interpolation=self.interpolation,
        )
        images = tf.ensure_shape(images, shape=original_shape)
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, raw_images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomAffine()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomAffine(bounding_box_format='xyxy')`"
            )
        # cast to float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            raw_images, dtype=tf.float32
        )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
            dtype=tf.float32,
        )
        boxes = bounding_boxes["boxes"]

        # process rotations
        origin_x = widths / 2
        origin_y = heights / 2
        angles = -transformations["angles"]
        angles = angles[:, tf.newaxis, tf.newaxis]
        # points: (batch_size, max_num_boxes, 4, 2)
        points = tf.stack(
            [
                tf.stack([boxes[:, :, 0], boxes[:, :, 1]], axis=2),
                tf.stack([boxes[:, :, 2], boxes[:, :, 1]], axis=2),
                tf.stack([boxes[:, :, 2], boxes[:, :, 3]], axis=2),
                tf.stack([boxes[:, :, 0], boxes[:, :, 3]], axis=2),
            ],
            axis=2,
        )
        point_x_offsets = points[..., 0] - origin_x[..., tf.newaxis]
        point_y_offsets = points[..., 1] - origin_y[..., tf.newaxis]
        new_x = (
            origin_x[..., tf.newaxis, tf.newaxis]
            + tf.multiply(tf.cos(angles), point_x_offsets[..., tf.newaxis])
            - tf.multiply(tf.sin(angles), point_y_offsets[..., tf.newaxis])
        )
        new_y = (
            origin_y[..., tf.newaxis, tf.newaxis]
            + tf.multiply(tf.sin(angles), point_x_offsets[..., tf.newaxis])
            + tf.multiply(tf.cos(angles), point_y_offsets[..., tf.newaxis])
        )
        out = tf.concat([new_x, new_y], axis=3)
        min_cordinates = tf.math.reduce_min(out, axis=2)
        max_cordinates = tf.math.reduce_max(out, axis=2)
        boxes = tf.concat([min_cordinates, max_cordinates], axis=2)

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="xyxy",
            images=raw_images,
        )
        # coordinates cannot be float values, it is cast to int32
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=raw_images,
        )
        return bounding_boxes

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        original_shape = segmentation_masks.shape
        batch_size = tf.shape(segmentation_masks)[0]
        rotation_matrixes = transformations["rotation_matrixes"]
        rotation_matrixes = tf.reshape(
            rotation_matrixes, shape=(batch_size, -1)
        )
        rotation_matrixes = rotation_matrixes[:, :-1]

        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if segmentation_masks.dtype == tf.bfloat16:
            segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        segmentation_masks = image_utils.transform(
            segmentation_masks,
            rotation_matrixes,
            fill_mode=self.fill_mode,
            fill_value=0,
            interpolation="nearest",
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=original_shape
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "factor": self.factor,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "interpolation": self.interpolation,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf

import keras_aug


class SignedNormalFactorSamplerTest(tf.test.TestCase):
    def test_sample(self):
        factor = keras_aug.SignedNormalFactorSampler(
            mean=0.5, stddev=0.2, min_value=0, max_value=1
        )

        sample = factor()

        self.assertTrue((0 <= sample <= 1) or (-1 <= sample <= 0))

    def test_sample_nostddev(self):
        factor = keras_aug.SignedNormalFactorSampler(
            mean=0.5, stddev=0.0, min_value=0, max_value=1
        )

        sample = factor()

        self.assertTrue((0 <= sample <= 1) or (-1 <= sample <= 0))

    def test_config(self):
        factor = keras_aug.SignedNormalFactorSampler(
            mean=0.5, stddev=0.2, min_value=0, max_value=1
        )

        config = factor.get_config()

        self.assertEqual(config["mean"], 0.5)
        self.assertEqual(config["stddev"], 0.2)

    def test_from_config(self):
        factor = keras_aug.SignedNormalFactorSampler(
            mean=0.5, stddev=0.2, min_value=0, max_value=1
        )
        config = factor.get_config()

        factor2 = keras_aug.SignedNormalFactorSampler.from_config(config)
        config = factor2.get_config()

        self.assertEqual(config["mean"], 0.5)
        self.assertEqual(config["stddev"], 0.2)



"""
Contains functions to compute ious of bounding boxes.

Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


def _compute_area(box):
    """Computes area for bounding boxes

    Args:
      box: [N, 4] or [batch_size, N, 4] float Tensor, either batched
        or unbatched boxes.
    Returns:
      a float Tensor of [N] or [batch_size, N]
    """
    y_min, x_min, y_max, x_max = tf.split(
        box[..., :4], num_or_size_splits=4, axis=-1
    )
    return tf.squeeze((y_max - y_min) * (x_max - x_min), axis=-1)


def _compute_intersection(boxes1, boxes2):
    """Computes intersection area between two sets of boxes.

    Args:
      boxes1: [N, 4] or [batch_size, N, 4] float Tensor boxes.
      boxes2: [M, 4] or [batch_size, M, 4] float Tensor boxes.
    Returns:
      a [N, M] or [batch_size, N, M] float Tensor.
    """
    y_min1, x_min1, y_max1, x_max1 = tf.split(
        boxes1[..., :4], num_or_size_splits=4, axis=-1
    )
    y_min2, x_min2, y_max2, x_max2 = tf.split(
        boxes2[..., :4], num_or_size_splits=4, axis=-1
    )
    boxes2_rank = len(boxes2.shape)
    perm = [1, 0] if boxes2_rank == 2 else [0, 2, 1]
    # [N, M] or [batch_size, N, M]
    intersect_ymax = tf.minimum(y_max1, tf.transpose(y_max2, perm))
    intersect_ymin = tf.maximum(y_min1, tf.transpose(y_min2, perm))
    intersect_xmax = tf.minimum(x_max1, tf.transpose(x_max2, perm))
    intersect_xmin = tf.maximum(x_min1, tf.transpose(x_min2, perm))

    intersect_height = intersect_ymax - intersect_ymin
    intersect_width = intersect_xmax - intersect_xmin
    zeros_t = tf.cast(0, intersect_height.dtype)
    intersect_height = tf.maximum(zeros_t, intersect_height)
    intersect_width = tf.maximum(zeros_t, intersect_width)

    return intersect_height * intersect_width


def compute_iou(
    boxes1,
    boxes2,
    bounding_box_format,
    use_masking=False,
    mask_val=-1,
    images=None,
    image_shape=None,
):
    """Computes a lookup table vector containing the ious for a given set boxes.

    The lookup vector is to be indexed by [`boxes1_index`,`boxes2_index`] if
    boxes are unbatched and by [`batch`, `boxes1_index`,`boxes2_index`] if the
    boxes are batched.

    The users can pass `boxes1` and `boxes2` to be different ranks. For example:
    1) `boxes1`: [batch_size, M, 4], `boxes2`: [batch_size, N, 4] -> return
        [batch_size, M, N].
    2) `boxes1`: [batch_size, M, 4], `boxes2`: [N, 4] -> return
        [batch_size, M, N]
    3) `boxes1`: [M, 4], `boxes2`: [batch_size, N, 4] -> return
        [batch_size, M, N]
    4) `boxes1`: [M, 4], `boxes2`: [N, 4] -> return [M, N]

    Args:
      boxes1: a list of bounding boxes in 'corners' format. Can be batched or
        unbatched.
      boxes2: a list of bounding boxes in 'corners' format. Can be batched or
        unbatched.
      bounding_box_format: The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
      use_masking: whether masking will be applied. This will mask all `boxes1`
        or `boxes2` that have values less than 0 in all its 4 dimensions.
        Default to `False`.
      mask_val: int to mask those returned IOUs if the masking is True, defaults
        to -1.

    Returns:
      iou_lookup_table: a vector containing the pairwise ious of boxes1 and
        boxes2.
    """  # noqa: E501

    boxes1_rank = len(boxes1.shape)
    boxes2_rank = len(boxes2.shape)

    if boxes1_rank not in [2, 3]:
        raise ValueError(
            "compute_iou() expects boxes1 to be batched, or to be unbatched. "
            f"Received len(boxes1.shape)={boxes1_rank}, "
            f"len(boxes2.shape)={boxes2_rank}. Expected either "
            "len(boxes1.shape)=2 AND or len(boxes1.shape)=3."
        )
    if boxes2_rank not in [2, 3]:
        raise ValueError(
            "compute_iou() expects boxes2 to be batched, or to be unbatched. "
            f"Received len(boxes1.shape)={boxes1_rank}, "
            f"len(boxes2.shape)={boxes2_rank}. Expected either "
            "len(boxes2.shape)=2 AND or len(boxes2.shape)=3."
        )

    target_format = "yxyx"
    if bounding_box.is_relative(bounding_box_format):
        target_format = bounding_box.as_relative(target_format)

    boxes1 = bounding_box.convert_format(
        boxes1,
        source=bounding_box_format,
        target=target_format,
        images=images,
        image_shape=image_shape,
    )

    boxes2 = bounding_box.convert_format(
        boxes2,
        source=bounding_box_format,
        target=target_format,
        images=images,
        image_shape=image_shape,
    )

    intersect_area = _compute_intersection(boxes1, boxes2)
    boxes1_area = _compute_area(boxes1)
    boxes2_area = _compute_area(boxes2)
    boxes2_area_rank = len(boxes2_area.shape)
    boxes2_axis = 1 if (boxes2_area_rank == 2) else 0
    boxes1_area = tf.expand_dims(boxes1_area, axis=-1)
    boxes2_area = tf.expand_dims(boxes2_area, axis=boxes2_axis)
    union_area = boxes1_area + boxes2_area - intersect_area
    res = tf.math.divide_no_nan(intersect_area, union_area)

    if boxes1_rank == 2:
        perm = [1, 0]
    else:
        perm = [0, 2, 1]

    if not use_masking:
        return res

    mask_val_t = tf.cast(mask_val, res.dtype) * tf.ones_like(res)
    boxes1_mask = tf.less(tf.reduce_max(boxes1, axis=-1, keepdims=True), 0.0)
    boxes2_mask = tf.less(tf.reduce_max(boxes2, axis=-1, keepdims=True), 0.0)
    background_mask = tf.logical_or(
        boxes1_mask, tf.transpose(boxes2_mask, perm)
    )
    iou_lookup_table = tf.where(background_mask, mask_val_t, res)
    return iou_lookup_table



import tensorflow as tf

from keras_aug import layers


class AugMixTest(tf.test.TestCase):
    def test_in_single_image(self):
        layer = layers.AugMix(value_range=(0, 255))

        # RGB
        xs = tf.cast(tf.ones((8, 8, 3)), dtype=tf.float32)

        xs = layer(xs)
        self.assertEqual(xs.shape, [8, 8, 3])

        # greyscale
        xs = tf.cast(tf.ones((8, 8, 1)), dtype=tf.float32)

        xs = layer(xs)
        self.assertEqual(xs.shape, [8, 8, 1])

    def test_non_square_images(self):
        layer = layers.AugMix(value_range=(0, 255))

        # RGB
        xs = tf.ones((2, 4, 8, 3))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 4, 8, 3])

        # greyscale
        xs = tf.ones((2, 4, 8, 1))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 4, 8, 1])



"""
Contains functions to compute ious of bounding boxes.

Most of these codes come from KerasCV.
"""
import tensorflow as tf

from keras_aug.datapoints import bounding_box


def _relative_area(boxes, bounding_box_format):
    boxes = bounding_box.convert_format(
        boxes,
        source=bounding_box_format,
        target="rel_xywh",
        dtype=boxes.dtype,
    )
    widths = boxes[..., 2]
    heights = boxes[..., 3]
    # handle corner case where shear performs a full inversion.
    return tf.where(
        tf.math.logical_and(widths > 0, heights > 0), widths * heights, 0.0
    )


def _compute_area(box):
    """Computes area for bounding boxes

    Args:
      box: [N, 4] or [batch_size, N, 4] float Tensor, either batched
        or unbatched boxes.
    Returns:
      a float Tensor of [N] or [batch_size, N]
    """
    y_min, x_min, y_max, x_max = tf.split(
        box[..., :4], num_or_size_splits=4, axis=-1
    )
    return tf.squeeze((y_max - y_min) * (x_max - x_min), axis=-1)


def _compute_intersection(boxes1, boxes2):
    """Computes intersection area between two sets of boxes.

    Args:
      boxes1: [N, 4] or [batch_size, N, 4] float Tensor boxes.
      boxes2: [M, 4] or [batch_size, M, 4] float Tensor boxes.
    Returns:
      a [N, M] or [batch_size, N, M] float Tensor.
    """
    y_min1, x_min1, y_max1, x_max1 = tf.split(
        boxes1[..., :4], num_or_size_splits=4, axis=-1
    )
    y_min2, x_min2, y_max2, x_max2 = tf.split(
        boxes2[..., :4], num_or_size_splits=4, axis=-1
    )
    boxes2_rank = len(boxes2.shape)
    perm = [1, 0] if boxes2_rank == 2 else [0, 2, 1]
    # [N, M] or [batch_size, N, M]
    intersect_ymax = tf.minimum(y_max1, tf.transpose(y_max2, perm))
    intersect_ymin = tf.maximum(y_min1, tf.transpose(y_min2, perm))
    intersect_xmax = tf.minimum(x_max1, tf.transpose(x_max2, perm))
    intersect_xmin = tf.maximum(x_min1, tf.transpose(x_min2, perm))

    intersect_height = intersect_ymax - intersect_ymin
    intersect_width = intersect_xmax - intersect_xmin
    zeros_t = tf.cast(0, intersect_height.dtype)
    intersect_height = tf.maximum(zeros_t, intersect_height)
    intersect_width = tf.maximum(zeros_t, intersect_width)

    return intersect_height * intersect_width


def _format_inputs(boxes, classes, images):
    boxes_rank = len(boxes.shape)
    if boxes_rank > 3:
        raise ValueError(
            "Expected len(boxes.shape)=2, or len(boxes.shape)=3, got "
            f"len(boxes.shape)={boxes_rank}"
        )
    boxes_includes_batch = boxes_rank == 3
    # Determine if images needs an expand_dims() call
    if images is not None:
        images_rank = len(images.shape)
        if images_rank > 4:
            raise ValueError(
                "Expected len(images.shape)=2, or len(images.shape)=3, got "
                f"len(images.shape)={images_rank}"
            )
        images_include_batch = images_rank == 4
        if boxes_includes_batch != images_include_batch:
            raise ValueError(
                "clip_to_image() expects both boxes and images to be batched, "
                "or both boxes and images to be unbatched. Received "
                f"len(boxes.shape)={boxes_rank}, "
                f"len(images.shape)={images_rank}. Expected either "
                "len(boxes.shape)=2 AND len(images.shape)=3, or "
                "len(boxes.shape)=3 AND len(images.shape)=4."
            )
        if not images_include_batch:
            images = tf.expand_dims(images, axis=0)

    if not boxes_includes_batch:
        return (
            tf.expand_dims(boxes, axis=0),
            tf.expand_dims(classes, axis=0),
            images,
            True,
        )
    return boxes, classes, images, False


def _format_outputs(boxes, classes, squeeze):
    if squeeze:
        return tf.squeeze(boxes, axis=0), tf.squeeze(classes, axis=0)
    return boxes, classes


def is_relative(bounding_box_format):
    """A util to check if a bounding box format uses relative coordinates"""
    if bounding_box_format.lower() not in bounding_box.TO_XYXY_CONVERTERS:
        raise ValueError(
            "`is_relative()` received an unsupported format for the argument "
            f"`bounding_box_format`. `bounding_box_format` should be one of "
            f"{bounding_box.TO_XYXY_CONVERTERS.keys()}. "
            f"Got bounding_box_format={bounding_box_format}"
        )

    return bounding_box_format.startswith("rel")


def as_relative(bounding_box_format):
    """A util to get the relative equivalent of a provided bounding box format.

    If the specified format is already a relative format,
    it will be returned unchanged.
    """

    if not is_relative(bounding_box_format):
        return "rel_" + bounding_box_format

    return bounding_box_format


def clip_to_image(
    bounding_boxes,
    bounding_box_format,
    images=None,
    image_shape=None,
):
    """clips bounding boxes to image boundaries.

    `clip_to_image()` clips bounding boxes that have coordinates out of bounds
    of an image down to the boundaries of the image. This is done by converting
    the bounding box to relative formats, then clipping them to the `[0, 1]`
    range. Additionally, bounding boxes that end up with a zero area have their
    class ID set to -1, indicating that there is no object present in them.

    Args:
        bounding_boxes: bounding box tensor to clip.
        bounding_box_format: the KerasCV bounding box format the bounding boxes
            are in.
        images: list of images to clip the bounding boxes to.
        image_shape: the shape of the images to clip the bounding boxes to.
    """
    boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]

    boxes = bounding_box.convert_format(
        boxes,
        source=bounding_box_format,
        target="rel_xyxy",
        images=images,
        image_shape=image_shape,
        dtype=boxes.dtype,
    )
    boxes, classes, images, squeeze = _format_inputs(boxes, classes, images)
    x1, y1, x2, y2 = tf.split(boxes, [1, 1, 1, 1], axis=-1)
    clipped_bounding_boxes = tf.concat(
        [
            tf.clip_by_value(x1, clip_value_min=0, clip_value_max=1),
            tf.clip_by_value(y1, clip_value_min=0, clip_value_max=1),
            tf.clip_by_value(x2, clip_value_min=0, clip_value_max=1),
            tf.clip_by_value(y2, clip_value_min=0, clip_value_max=1),
        ],
        axis=-1,
    )
    areas = _relative_area(
        clipped_bounding_boxes, bounding_box_format="rel_xyxy"
    )
    clipped_bounding_boxes = bounding_box.convert_format(
        clipped_bounding_boxes,
        source="rel_xyxy",
        target=bounding_box_format,
        images=images,
        image_shape=image_shape,
        dtype=clipped_bounding_boxes.dtype,
    )
    clipped_bounding_boxes = tf.where(
        tf.expand_dims(areas > 0.0, axis=-1),
        clipped_bounding_boxes,
        tf.constant(-1, dtype=clipped_bounding_boxes.dtype),
    )
    classes = tf.where(areas > 0.0, classes, tf.constant(-1, classes.dtype))
    nan_indices = tf.math.reduce_any(
        tf.math.is_nan(clipped_bounding_boxes), axis=-1
    )
    classes = tf.where(nan_indices, tf.constant(-1, classes.dtype), classes)

    # TODO update dict and return
    clipped_bounding_boxes, classes = _format_outputs(
        clipped_bounding_boxes, classes, squeeze
    )

    result = bounding_boxes.copy()
    result["boxes"] = clipped_bounding_boxes
    result["classes"] = classes
    return result


def sanitize_bounding_boxes(
    bounding_boxes,
    min_size=None,
    min_area_ratio=None,
    max_aspect_ratio=None,
    bounding_box_format=None,
    reference_bounding_boxes=None,
    images=None,
    reference_images=None,
):
    """Sanitize bounding boxes by min_size, min_area_ratio and max_aspect_ratio.

    Args:
        bounding_boxes (dict[str, tf.Tensor]): The bounding boxes to sanitize.
        min_size (float, optional): The minimum size of the bounding boxes.
        min_area_ratio (float, optional): The minimum area ratio of original
            bounding boxes to bounding boxes for sanitizing.
            Defaults to ``None``.
        max_aspect_ratio (float, optional): The maximum aspect ratio of
            bounding boxes for sanitizing. Defaults to ``None``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        reference_bounding_boxes (dict[str, tf.Tensor], optional): The
            reference bounding boxes when apply sanitizing with
            min_area_ratio enabled. Defaults to ``None``.
        images (tf.Tensor, optional): The images for bounding boxes
            format conversion.
        reference_images (tf.Tensor, optional): The reference images for
            reference bounding boxes format conversion.

    References:
        - `ultralytics/ultralytics <https://github.com/ultralytics/ultralytics>`_
    """  # noqa: E501
    if min_size is None and min_area_ratio is None and max_aspect_ratio is None:
        return bounding_boxes
    if min_size is not None:
        assert isinstance(min_size, (int, float))
    if min_area_ratio is not None:
        assert isinstance(min_area_ratio, (int, float))
    if max_aspect_ratio is not None:
        assert isinstance(max_aspect_ratio, (int, float))

    sanitize_mask = tf.zeros(
        tf.shape(bounding_boxes["boxes"])[:2], dtype=tf.bool
    )

    if min_size is not None:
        if bounding_box_format is None or images is None:
            raise ValueError(
                "When apply sanitize_bounding_boxes with "
                "min_size, must pass bounding_box_format and images."
            )
        boxes = bounding_boxes["boxes"]
        boxes = bounding_box.convert_format(
            boxes,
            source=bounding_box_format,
            target="xywh",
            images=images,
            dtype=boxes.dtype,
        )
        _, _, widths, heights = tf.split(boxes, 4, axis=-1)
        min_sides = tf.minimum(widths, heights)
        min_sides = tf.squeeze(min_sides, axis=-1)
        sanitize_mask = tf.math.logical_or(sanitize_mask, min_sides < min_size)

    if min_area_ratio is not None:
        if (
            bounding_box_format is None
            or reference_bounding_boxes is None
            or images is None
            or reference_images is None
        ):
            raise ValueError(
                "When apply sanitize_bounding_boxes with "
                "min_area_ratio, must pass bounding_box_format, "
                "reference_bounding_boxes, images and reference_images."
                "."
            )
        ref_boxes = reference_bounding_boxes["boxes"]
        ref_boxes = bounding_box.convert_format(
            ref_boxes,
            source=bounding_box_format,
            target="xyxy",
            images=reference_images,
            dtype=ref_boxes.dtype,
        )
        boxes = bounding_boxes["boxes"]
        boxes = bounding_box.convert_format(
            boxes,
            source=bounding_box_format,
            target="xyxy",
            images=images,
            dtype=boxes.dtype,
        )
        ref_areas = _compute_area(ref_boxes)
        areas = _compute_area(boxes)
        area_ratios = tf.math.divide_no_nan(areas, ref_areas)
        sanitize_mask = tf.math.logical_or(
            sanitize_mask, area_ratios < min_area_ratio
        )

    if max_aspect_ratio is not None:
        if bounding_box_format is None or images is None:
            raise ValueError(
                "When apply sanitize_bounding_boxes with "
                "max_aspect_ratio, must pass bounding_box_format and "
                "images."
            )
        boxes = bounding_boxes["boxes"]
        boxes = bounding_box.convert_format(
            boxes,
            source=bounding_box_format,
            target="xywh",
            images=images,
            dtype=boxes.dtype,
        )
        _, _, widths, heights = tf.split(boxes, 4, axis=-1)
        max_aspect_ratios = tf.squeeze(
            tf.maximum(
                tf.math.divide_no_nan(widths, heights),
                tf.math.divide_no_nan(heights, widths),
            ),
            axis=-1,
        )
        sanitize_mask = tf.math.logical_or(
            sanitize_mask, max_aspect_ratios > max_aspect_ratio
        )

    # set classes == -1
    bounding_boxes = bounding_boxes.copy()
    bounding_boxes["classes"] = tf.where(
        sanitize_mask,
        tf.constant(-1, bounding_boxes["classes"].dtype),
        bounding_boxes["classes"],
    )
    return bounding_boxes



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils.augmentation import BATCHED
from keras_aug.utils.distribution import stateless_random_beta


@keras.utils.register_keras_serializable(package="keras_aug")
class MixUp(VectorizedBaseRandomLayer):
    """The MixUp data augmentation technique.

    The MixUp data augmentation technique involves taking 2 images from a given
    batch and fusing them together using a ratio sampled from a beta
    distribution. Labels are applied by same ratio ratio. Bounding boxes are
    concated according to the position of the 2 images.

    Args:
        alpha (float, optional): The inverse scale parameter between 0 to +inf
            for the gamma distribution. This controls the shape of the
            distribution from which the smoothing values are sampled. Defaults
            to ``0.2``, which is a recommended value when training an ImageNet
            classification model. For object detection, it is recommended to use
            a larger value. For example YOLOV8 uses ``32.0``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `MixUp <https://arxiv.org/abs/1710.09412>`_
        - `Bag of Freebies for Training Object Detection Neural Networks <https://arxiv.org/abs/1902.04103>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(self, alpha=0.2, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.alpha = alpha
        self.seed = seed

        # set force_no_unwrap_ragged_image_call=True because MixUp needs
        # to process images in batch.
        self.force_no_unwrap_ragged_image_call = True

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # cast to float32 to avoid numerical issue
        permutation_order = tf.argsort(
            self._random_generator.random_uniform((batch_size,)), axis=-1
        )
        lambda_samples = stateless_random_beta(
            (batch_size, 1, 1, 1),
            seed_alpha=self._random_generator.make_seed_for_stateless_op(),
            seed_beta=self._random_generator.make_seed_for_stateless_op(),
            alpha=self.alpha,
            beta=self.alpha,
            dtype=tf.float32,
        )
        return {
            "permutation_order": permutation_order,
            "lambda_samples": lambda_samples,
        }

    def augment_images(self, images, transformations, **kwargs):
        if isinstance(images, tf.RaggedTensor):
            raise ValueError(
                "MixUp expects dense images. Received: images type: "
                f"{type(images)}"
            )
        permutation_order = transformations["permutation_order"]
        lambda_samples = tf.cast(
            transformations["lambda_samples"], dtype=self.compute_dtype
        )
        mixup_images = tf.gather(images, permutation_order)
        mixup_images = tf.cast(mixup_images, dtype=self.compute_dtype)
        images = lambda_samples * images + (1.0 - lambda_samples) * mixup_images
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        permutation_order = transformations["permutation_order"]
        lambda_samples = tf.cast(
            transformations["lambda_samples"], dtype=self.compute_dtype
        )
        labels = tf.cast(labels, dtype=self.compute_dtype)
        mixup_labels = tf.gather(labels, permutation_order)
        lambda_samples = tf.reshape(lambda_samples, [-1, 1])
        labels = lambda_samples * labels + (1.0 - lambda_samples) * mixup_labels
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        permutation_order = transformations["permutation_order"]
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]
        mixup_boxes = tf.gather(boxes, permutation_order)
        mixup_classes = tf.gather(classes, permutation_order)
        boxes = tf.concat([boxes, mixup_boxes], axis=1)
        classes = tf.concat([classes, mixup_classes], axis=1)

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes["classes"] = classes
        return bounding_boxes

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        return super()._batch_augment(inputs)

    def call(self, inputs):
        _, metadata = self._format_inputs(inputs)
        if metadata[BATCHED] is not True:
            raise ValueError(
                "MixUp received a single image to `call`. The "
                "layer relies on combining multiple examples, and as such "
                "will not behave as expected. Please call the layer with 2 "
                "or more samples."
            )
        return super().call(inputs=inputs)

    def _validate_inputs(self, inputs):
        images = inputs.get("images", None)
        labels = inputs.get("labels", None)
        bounding_boxes = inputs.get("bounding_boxes", None)
        if images is None or (labels is None and bounding_boxes is None):
            raise ValueError(
                "MixUp expects inputs in a dictionary with format "
                '{"images": images, "labels": labels}. or'
                '{"images": images, "bounding_boxes": bounding_boxes}'
                f"Got: inputs = {inputs}."
            )
        if bounding_boxes is not None:
            _ = bounding_box.validate_format(bounding_boxes)

    def sample_from_beta(self, alpha, beta, shape):
        sample_alpha = tf.random.stateless_gamma(
            shape,
            alpha=alpha,
            seed=self._random_generator.make_seed_for_stateless_op(),
        )
        sample_beta = tf.random.stateless_gamma(
            shape,
            alpha=beta,
            seed=self._random_generator.make_seed_for_stateless_op(),
        )
        return sample_alpha / (sample_alpha + sample_beta)

    def get_config(self):
        config = super().get_config()
        config.update({"alpha": self.alpha, "seed": self.seed})
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.__internal__.base_layer import BaseRandomLayer
from keras_aug.utils import augmentation as augmentation_utils

H_AXIS = -3
W_AXIS = -2

IMAGES = "images"
LABELS = "labels"
TARGETS = "targets"
BOUNDING_BOXES = "bounding_boxes"
KEYPOINTS = "keypoints"
SEGMENTATION_MASKS = "segmentation_masks"
CUSTOM_ANNOTATIONS = "custom_annotations"

IS_DICT = "is_dict"
BATCHED = "batched"
USE_TARGETS = "use_targets"


@keras.utils.register_keras_serializable(package="keras_aug")
class VectorizedBaseRandomLayer(BaseRandomLayer):
    """Abstract base layer for vectorized image augmentation.

    This layer contains base functionalities for preprocessing layers which
    augment image related data, e.g. image and in the future, label and bounding
    boxes. The subclasses could avoid making certain mistakes and reduce code
    duplications.

    This layer requires you to implement one method ``augment_images()``, which
    augments one single image during the training. There are a few additional
    methods that you can implement for added functionality on the layer.

    ``augment_ragged_image()`` and ``compute_ragged_image_signature()``, which
    handles ragged images augmentation if the layer supports that.

    ``augment_labels()``, which handles label augmentation if the layer
    supports that.

    ``augment_bounding_boxes()``, which handles the bounding box
    augmentation, if the layer supports that.

    ``augment_keypoints()``, which handles the keypoints augmentation, if the
    layer supports that.

    ``augment_segmentation_masks()``, which handles the segmentation masks
    augmentation, if the layer supports that.

    ``augment_custom_annotations()``, which handles the custom annotations
    augmentation, if the layer supports that. This is useful to implement
    augmentation for special annotatinos.

    ``get_random_transformations()``, which should produce a batch of random
    transformation settings. The transformation object, which must be a
    batched Tensor or a dictionary where each input is a batched Tensor,
    will be passed to ``augment_images``, ``augment_labels`` and
    `augment_bounding_boxes`, to coordinate the randomness behavior, eg, in
    the RandomFlip layer, the image and bounding_boxes should be changed in
    the same way.

    The ``call()`` method support two formats of inputs::

        1. Single image tensor with 3D (HWC) or 4D (NHWC) format.
        2. A dict of tensors with stable keys. The supported keys are
            ``"images"``, ``"labels"``, ``"bounding_boxes"``,
            ``segmentation_masks``, ``keypoints`` and ``custom_annotations`` at
            the moment. We might add more keys in future when we support more
            types of augmentation.

    The output of the ``call()`` will be in two formats, which will be the same
    structure as the inputs.

    The ``call()`` will unpack the inputs, forward to the correct function, and
    pack the output back to the same structure as the inputs.

    By default, the dense or ragged status of the output will be preserved.
    However, you can override this behavior by setting
    ``self.force_output_dense_images = True`` in your ``__init__()``
    method. When enabled, images and segmentation masks will be converted to
    dense tensor by ``to_tensor()`` if ragged.

    .. code-block:: python

        class SubclassLayer(VectorizedBaseImageAugmentationLayer):
            def __init__(self):
                super().__init__()
                self.force_output_dense_images = True

    Note that since the randomness is also a common functionality, this layer
    also includes a keras.src.backend.RandomGenerator, which can be used to
    produce the random numbers. The random number generator is stored in the
    `self._random_generator` attribute.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(
        self, seed=None, force_generator=True, rng_type="stateless", **kwargs
    ):
        # Defaults to stateless random operations
        super().__init__(
            seed=seed,
            force_generator=force_generator,
            rng_type=rng_type,
            **kwargs,
        )

    @property
    def force_no_unwrap_ragged_image_call(self):
        """Control whether to force not to unwrap ragged image call."""
        return getattr(self, "_force_no_unwrap_ragged_image_call", False)

    @force_no_unwrap_ragged_image_call.setter
    def force_no_unwrap_ragged_image_call(
        self, force_no_unwrap_ragged_image_call
    ):
        self._force_no_unwrap_ragged_image_call = (
            force_no_unwrap_ragged_image_call
        )

    @property
    def force_output_dense_images(self):
        """Control whether to force outputting of dense images."""
        return getattr(self, "_force_output_dense_images", False)

    @force_output_dense_images.setter
    def force_output_dense_images(self, force_output_dense_images):
        self._force_output_dense_images = force_output_dense_images

    def get_random_transformation_batch(
        self,
        batch_size,
        images=None,
        labels=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_masks=None,
        custom_annotations=None,
    ):
        """Produce random transformations config for a batch of inputs.

        This is used to produce same randomness between image / label /
        bounding_box.

        Args:
            batch_size: the batch size of transformations configuration to
                sample.
            images: 3D image tensor from inputs.
            labels: optional 1D label tensor from inputs.
            bounding_boxes: optional 2D bounding boxes tensor from inputs.
            segmentation_masks: optional 3D segmentation mask tensor from
                inputs.

        Returns:
            Any type of object, which will be forwarded to `augment_images`,
            `augment_labels` and `augment_bounding_boxes` as the
            `transformations` parameter.
        """
        # Required to work with map_fn in the ragged cast.
        return tf.zeros((batch_size))

    def compute_ragged_image_signature(self, images):
        """Computes the output image signature for the
        `_unwrap_ragged_image_call()` function.

        Must be overridden to return tensors with different shapes than the
        input images. By default, returns either a `tf.RaggedTensorSpec`
        matching the input image spec, or a `tf.TensorSpec` matching the input
        image spec.
        """
        return tf.RaggedTensorSpec(
            shape=images.shape[1:],
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def compute_ragged_segmentation_mask_signature(self, segmentation_maks):
        """Computes the output segmentation_mask signature for the
        `_unwrap_ragged_segmentation_call()` function.

        Must be overridden to return tensors with different shapes than the
        input images. By default, returns either a `tf.RaggedTensorSpec`
        matching the input image spec, or a `tf.TensorSpec` matching the input
        image spec.
        """
        return tf.RaggedTensorSpec(
            shape=segmentation_maks.shape[1:],
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        """Augment an image from a ragged image batch.

        This method accepts a single Dense image Tensor, and returns a Dense
        image. The resulting images are then stacked back into a ragged image
        batch. The behavior of this method should be identical to that of
        `augment_images()` but is to operate on a batch-wise basis.

        Args:
            image: a single image from the batch
            transformation: a single transformation sampled from
                `get_random_transformations()`.
            kwargs: all the other call arguments (i.e. bounding_boxes, labels,
                etc.).

        Returns:
            Augmented image.
        """
        raise NotImplementedError(
            "A ragged image batch was passed to layer of type "
            f"`{type(self).__name__}`. This layer does not implement "
            "`augment_ragged_image()`. If this is a `keras_aug`, open a GitHub "
            "issue requesting Ragged functionality on the layer titled: "
            f"'`{type(self).__name__}`: ragged image support'. If this is a "
            "custom layer, implement the `augment_ragged_image()` method."
        )

    def augment_images(self, images, transformations, **kwargs):
        """Augment a batch of images.

        Args:
            images: 4D image input tensor to the layer. Forwarded from
                `layer.call()`. This should generally have the shape
                [B, H, W, C]. Forwarded from `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_labels(self, labels, transformations, **kwargs):
        """Augment a batch of labels.

        Args:
            labels: 2D label to the layer. Forwarded from `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 2D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        """Augment bounding boxes for one image.

        Args:
            bounding_boxes: 3D bounding boxes to the layer. Forwarded from
                `call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        """Augment a batch of keypoints for one image.

        Args:
            keypoints: 3D keypoints input tensor to the layer. Forwarded from
                `layer.call()`. Shape should be [batch, num_keypoints, 2] in the
                specified keypoint format.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        """Augment an image from a ragged segmentation mask batch.

        This method accepts a single Dense image Tensor, and returns a Dense
        image. The resulting images are then stacked back into a ragged image
        batch. The behavior of this method should be identical to that of
        `augment_segmentation_masks()` but is to operate on a batch-wise basis.

        Args:
            segmentation_mask: a single image from the batch
            transformation: a single transformation sampled from
                `get_random_transformations()`.
            kwargs: all the other call arguments (i.e. bounding_boxes, labels,
                etc.).

        Returns:
            Augmented segmentation mask.
        """
        return segmentation_mask

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        """Augment a batch of images' segmentation masks.

        Args:
            segmentation_masks: 4D segmentation mask input tensor to the layer.
                This should generally have the shape [B, H, W, 1], or in some
                cases [B, H, W, C] for multilabeled data. Forwarded from
                `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor containing the augmented segmentation mask, which
            will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_custom_annotations(
        self, custom_annotations, transformations, **kwargs
    ):
        """Augment a batch of images' custom_annotations.

        Args:
            custom_annotations: 4D custom annotations input tensor to the layer.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor containing the augmented segmentation mask, which
            will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def _unwrap_ragged_image_call(self, inputs):
        images = inputs.get(IMAGES, None)
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        keypoints = inputs.get(KEYPOINTS, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        transformations = inputs.get("transformations")
        images = images.to_tensor()
        images = self.augment_ragged_image(
            image=images,
            label=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_mask=segmentation_masks,
            transformation=transformations,
        )
        return tf.RaggedTensor.from_tensor(images)

    def _unwrap_ragged_segmentation_mask_call(self, inputs):
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        transformations = inputs.get("transformations")
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        images = inputs.get(IMAGES, None)
        raw_images = inputs.get("raw_images", None)
        segmentation_masks = segmentation_masks.to_tensor()
        segmentation_masks = self.augment_ragged_segmentation_mask(
            segmentation_mask=segmentation_masks,
            transformation=transformations,
            label=labels,
            bounding_boxes=bounding_boxes,
            image=images,
            raw_image=raw_images,
        )
        return tf.RaggedTensor.from_tensor(segmentation_masks)

    def _batch_augment(self, inputs):
        images = inputs.get(IMAGES, None)
        raw_images = images
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        keypoints = inputs.get(KEYPOINTS, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        custom_annotations = inputs.get(CUSTOM_ANNOTATIONS, None)
        batch_size = tf.shape(images)[0]

        transformations = self.get_random_transformation_batch(
            batch_size,
            images=images,
            labels=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_masks=segmentation_masks,
            custom_annotations=custom_annotations,
        )

        if (
            isinstance(images, tf.RaggedTensor)
            and not self.force_no_unwrap_ragged_image_call
        ):
            inputs_for_raggeds = {"transformations": transformations, **inputs}
            images = tf.map_fn(
                self._unwrap_ragged_image_call,
                inputs_for_raggeds,
                fn_output_signature=self.compute_ragged_image_signature(images),
            )
        else:
            images = self.augment_images(
                images,
                transformations=transformations,
                bounding_boxes=bounding_boxes,
                labels=labels,
            )
        if (
            isinstance(images, tf.RaggedTensor)
            and self.force_output_dense_images
        ):
            images = images.to_tensor()
        result = {IMAGES: images}

        if labels is not None:
            labels = self.augment_labels(
                labels,
                transformations=transformations,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[LABELS] = labels

        if bounding_boxes is not None:
            ori_bounding_boxes_info = bounding_box.validate_format(
                bounding_boxes
            )
            bounding_boxes = self.augment_bounding_boxes(
                bounding_boxes,
                transformations=transformations,
                labels=labels,
                images=images,
                raw_images=raw_images,
            )
            if ori_bounding_boxes_info["ragged"]:
                bounding_boxes = bounding_box.to_ragged(
                    bounding_boxes, dtype=self.compute_dtype
                )
            result[BOUNDING_BOXES] = bounding_boxes

        if keypoints is not None:
            keypoints = self.augment_keypoints(
                keypoints,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[KEYPOINTS] = keypoints

        if segmentation_masks is not None:
            if (
                isinstance(segmentation_masks, tf.RaggedTensor)
                and not self.force_no_unwrap_ragged_image_call
            ):
                inputs_for_raggeds = {
                    "transformations": transformations,
                    "raw_images": raw_images,
                    **inputs,
                }
                segmentation_masks = tf.map_fn(
                    self._unwrap_ragged_segmentation_mask_call,
                    inputs_for_raggeds,
                    fn_output_signature=self.compute_ragged_segmentation_mask_signature(  # noqa: E501
                        segmentation_masks
                    ),
                )
            else:
                segmentation_masks = self.augment_segmentation_masks(
                    segmentation_masks,
                    transformations=transformations,
                    labels=labels,
                    bounding_boxes=bounding_boxes,
                    images=images,
                    raw_images=raw_images,
                )
            if (
                isinstance(segmentation_masks, tf.RaggedTensor)
                and self.force_output_dense_images
            ):
                segmentation_masks = segmentation_masks.to_tensor()
            result[SEGMENTATION_MASKS] = segmentation_masks

        if custom_annotations is not None:
            custom_annotations = self.augment_custom_annotations(
                custom_annotations,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[CUSTOM_ANNOTATIONS] = custom_annotations

        # preserve any additional inputs unmodified by this layer.
        for key in inputs.keys() - result.keys():
            result[key] = inputs[key]
        return result

    def _ensure_inputs_are_compute_dtype(self, inputs):
        if not isinstance(inputs, dict):
            return augmentation_utils.ensure_tensor(
                inputs,
                self.compute_dtype,
            )
        inputs[IMAGES] = augmentation_utils.ensure_tensor(
            inputs[IMAGES],
            self.compute_dtype,
        )
        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES]["boxes"] = augmentation_utils.ensure_tensor(
                inputs[BOUNDING_BOXES]["boxes"],
                self.compute_dtype,
            )
            inputs[BOUNDING_BOXES][
                "classes"
            ] = augmentation_utils.ensure_tensor(
                inputs[BOUNDING_BOXES]["classes"],
                self.compute_dtype,
            )
        return inputs

    def _format_inputs(self, inputs):
        metadata = {IS_DICT: True, USE_TARGETS: False}
        if tf.is_tensor(inputs):
            metadata[IS_DICT] = False
            inputs = {IMAGES: inputs}

        metadata[BATCHED] = inputs["images"].shape.rank == 4
        if inputs["images"].shape.rank == 3:
            for key in list(inputs.keys()):
                if key == BOUNDING_BOXES:
                    inputs[BOUNDING_BOXES]["boxes"] = tf.expand_dims(
                        inputs[BOUNDING_BOXES]["boxes"], axis=0
                    )
                    inputs[BOUNDING_BOXES]["classes"] = tf.expand_dims(
                        inputs[BOUNDING_BOXES]["classes"], axis=0
                    )
                else:
                    inputs[key] = tf.expand_dims(inputs[key], axis=0)

        if not isinstance(inputs, dict):
            raise ValueError(
                "Expect the inputs to be image tensor or dict. Got "
                f"inputs={inputs}"
            )

        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES] = self._format_bounding_boxes(
                inputs[BOUNDING_BOXES]
            )

        if isinstance(inputs, dict) and TARGETS in inputs:
            inputs[LABELS] = inputs[TARGETS]
            del inputs[TARGETS]
            metadata[USE_TARGETS] = True
            return inputs, metadata

        return inputs, metadata

    def _format_bounding_boxes(self, bounding_boxes):
        # We can't catch the case where this is None, sometimes RaggedTensor
        # drops this dimension.
        if "classes" not in bounding_boxes:
            raise ValueError(
                "Bounding boxes are missing class_id. If you would like to pad "
                "the bounding boxes with class_id, use: "
                "`bounding_boxes['classes'] = "
                "tf.ones_like(bounding_boxes['boxes'])`."
            )
        return bounding_boxes

    def _format_output(self, output, metadata):
        if not metadata[BATCHED]:
            for key in list(output.keys()):
                if key == BOUNDING_BOXES:
                    output[BOUNDING_BOXES]["boxes"] = tf.squeeze(
                        output[BOUNDING_BOXES]["boxes"], axis=0
                    )
                    output[BOUNDING_BOXES]["classes"] = tf.squeeze(
                        output[BOUNDING_BOXES]["classes"], axis=0
                    )
                else:
                    output[key] = tf.squeeze(output[key], axis=0)

        if not metadata[IS_DICT]:
            return output[IMAGES]
        elif metadata[USE_TARGETS]:
            output[TARGETS] = output[LABELS]
            del output[LABELS]
        return output

    def call(self, inputs):
        inputs = self._ensure_inputs_are_compute_dtype(inputs)
        inputs, metadata = self._format_inputs(inputs)
        images = inputs[IMAGES]
        if images.shape.rank == 3 or images.shape.rank == 4:
            return self._format_output(self._batch_augment(inputs), metadata)
        else:
            raise ValueError(
                "Image augmentation layers are expecting inputs to be "
                "rank 3 (HWC) or 4D (NHWC) tensors. Got shape: "
                f"{images.shape}"
            )

    @classmethod
    def from_config(cls, config):
        return cls(**config)



import numpy as np
import tensorflow as tf

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


class VectorizedRandomAddLayer(VectorizedBaseRandomLayer):
    def __init__(self, add_range=(0.0, 1.0), fixed_value=None, **kwargs):
        super().__init__(**kwargs)
        self.add_range = add_range
        self.fixed_value = fixed_value

    def get_random_transformation_batch(self, batch_size, **kwargs):
        if self.fixed_value:
            return tf.ones((batch_size,)) * self.fixed_value
        return self._random_generator.random_uniform(
            (batch_size,), minval=self.add_range[0], maxval=self.add_range[1]
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        return image + transformation[None, None, None]

    def augment_images(self, images, transformations, **kwargs):
        return images + transformations[:, None, None, None]

    def augment_labels(self, labels, transformations, **kwargs):
        return labels + transformations[:, None]

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return {
            "boxes": bounding_boxes["boxes"] + transformations[:, None, None],
            "classes": bounding_boxes["classes"] + transformations[:, None],
        }

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints + transformations[:, None, None]

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        return segmentation_mask + transformation[None, None, None]

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks + transformations[:, None, None, None]


TF_ALL_TENSOR_TYPES = (tf.Tensor, tf.RaggedTensor, tf.SparseTensor)


class VectorizedAssertionLayer(VectorizedBaseRandomLayer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_random_transformation_batch(
        self,
        batch_size,
        images=None,
        labels=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_masks=None,
        **kwargs,
    ):
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(keypoints, TF_ALL_TENSOR_TYPES)
        assert isinstance(segmentation_masks, TF_ALL_TENSOR_TYPES)
        return self._random_generator.random_uniform((batch_size,))

    def augment_ragged_image(
        self,
        image,
        label=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_mask=None,
        transformation=None,
        **kwargs,
    ):
        assert isinstance(image, TF_ALL_TENSOR_TYPES)
        assert isinstance(label, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(keypoints, TF_ALL_TENSOR_TYPES)
        assert isinstance(segmentation_mask, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformation, TF_ALL_TENSOR_TYPES)
        return image

    def augment_images(
        self,
        images,
        transformations=None,
        bounding_boxes=None,
        labels=None,
        **kwargs,
    ):
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        return images

    def augment_labels(
        self,
        labels,
        transformations=None,
        bounding_boxes=None,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_images, TF_ALL_TENSOR_TYPES)
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations=None,
        labels=None,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_images, TF_ALL_TENSOR_TYPES)
        return bounding_boxes

    def augment_keypoints(
        self,
        keypoints,
        transformations=None,
        labels=None,
        bounding_boxes=None,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        assert isinstance(keypoints, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_images, TF_ALL_TENSOR_TYPES)
        return keypoints

    def augment_ragged_segmentation_mask(
        self,
        segmentation_mask,
        transformation=None,
        label=None,
        bounding_boxes=None,
        image=None,
        raw_image=None,
        **kwargs,
    ):
        assert isinstance(segmentation_mask, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformation, TF_ALL_TENSOR_TYPES)
        assert isinstance(label, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(image, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_image, TF_ALL_TENSOR_TYPES)
        return segmentation_mask

    def augment_segmentation_masks(
        self,
        segmentation_masks,
        transformations=None,
        labels=None,
        bounding_boxes=None,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        assert isinstance(segmentation_masks, TF_ALL_TENSOR_TYPES)
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_images, TF_ALL_TENSOR_TYPES)
        return segmentation_masks

    def augment_custom_annotations(
        self,
        custom_annotations,
        transformations=None,
        labels=None,
        bounding_boxes=None,
        images=None,
        raw_images=None,
    ):
        assert isinstance(transformations, TF_ALL_TENSOR_TYPES)
        assert isinstance(labels, TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["boxes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(bounding_boxes["classes"], TF_ALL_TENSOR_TYPES)
        assert isinstance(images, TF_ALL_TENSOR_TYPES)
        assert isinstance(raw_images, TF_ALL_TENSOR_TYPES)
        return custom_annotations


class VectorizedBaseRandomLayerTest(tf.test.TestCase):
    def test_augment_single_image(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        image = np.random.random(size=(8, 8, 3)).astype("float32")
        output = add_layer(image)

        self.assertAllClose(image + 2.0, output)

    def test_augment_dict_return_type(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        image = np.random.random(size=(8, 8, 3)).astype("float32")
        output = add_layer({"images": image})

        self.assertIsInstance(output, dict)

    def test_augment_casts_dtypes(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        images = tf.ones((2, 8, 8, 3), dtype="uint8")
        output = add_layer(images)

        self.assertAllClose(
            tf.ones((2, 8, 8, 3), dtype="float32") * 3.0, output
        )

    def test_augment_batch_images(self):
        add_layer = VectorizedRandomAddLayer()
        images = np.random.random(size=(2, 8, 8, 3)).astype("float32")
        output = add_layer(images)

        diff = output - images
        # Make sure the first image and second image get different augmentation
        self.assertNotAllClose(diff[0], diff[1])

    def test_augment_image_and_label(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        image = np.random.random(size=(8, 8, 3)).astype("float32")
        label = np.random.random(size=(1,)).astype("float32")

        output = add_layer({"images": image, "targets": label})
        expected_output = {"images": image + 2.0, "targets": label + 2.0}
        self.assertAllClose(output, expected_output)

    def test_augment_image_and_target(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        image = np.random.random(size=(8, 8, 3)).astype("float32")
        label = np.random.random(size=(1,)).astype("float32")

        output = add_layer({"images": image, "targets": label})
        expected_output = {"images": image + 2.0, "targets": label + 2.0}
        self.assertAllClose(output, expected_output)

    def test_augment_batch_images_and_targets(self):
        add_layer = VectorizedRandomAddLayer()
        images = np.random.random(size=(2, 8, 8, 3)).astype("float32")
        targets = np.random.random(size=(2, 1)).astype("float32")
        output = add_layer({"images": images, "targets": targets})

        image_diff = output["images"] - images
        label_diff = output["targets"] - targets
        # Make sure the first image and second image get different augmentation
        self.assertNotAllClose(image_diff[0], image_diff[1])
        self.assertNotAllClose(label_diff[0], label_diff[1])

    def test_augment_leaves_extra_dict_entries_unmodified(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=0.5)
        images = np.random.random(size=(8, 8, 3)).astype("float32")
        filenames = tf.constant("/path/to/first.jpg")
        inputs = {"images": images, "filenames": filenames}
        output = add_layer(inputs)
        self.assertAllEqual(output["filenames"], filenames)

    def test_augment_ragged_images(self):
        images = tf.ragged.stack(
            [
                np.random.random(size=(8, 8, 3)).astype("float32"),
                np.random.random(size=(16, 8, 3)).astype("float32"),
            ]
        )
        add_layer = VectorizedRandomAddLayer(fixed_value=0.5)
        result = add_layer(images)
        self.assertAllClose(images + 0.5, result)

    def test_augment_image_and_localization_data(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        images = np.random.random(size=(8, 8, 8, 3)).astype("float32")
        bounding_boxes = {
            "boxes": np.random.random(size=(8, 3, 4)).astype("float32"),
            "classes": np.random.random(size=(8, 3)).astype("float32"),
        }
        keypoints = np.random.random(size=(8, 5, 2)).astype("float32")
        segmentation_mask = np.random.random(size=(8, 8, 8, 1)).astype(
            "float32"
        )

        output = add_layer(
            {
                "images": images,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_mask,
            }
        )
        expected_output = {
            "images": images + 2.0,
            "bounding_boxes": bounding_box.to_dense(
                {
                    "boxes": bounding_boxes["boxes"] + 2.0,
                    "classes": bounding_boxes["classes"] + 2.0,
                }
            ),
            "keypoints": keypoints + 2.0,
            "segmentation_masks": segmentation_mask + 2.0,
        }

        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(output["images"], expected_output["images"])
        self.assertAllClose(output["keypoints"], expected_output["keypoints"])
        self.assertAllClose(
            output["bounding_boxes"]["boxes"],
            expected_output["bounding_boxes"]["boxes"],
        )
        self.assertAllClose(
            output["bounding_boxes"]["classes"],
            expected_output["bounding_boxes"]["classes"],
        )
        self.assertAllClose(
            output["segmentation_masks"], expected_output["segmentation_masks"]
        )

    def test_augment_batch_image_and_localization_data(self):
        add_layer = VectorizedRandomAddLayer()
        images = np.random.random(size=(2, 8, 8, 3)).astype("float32")
        bounding_boxes = {
            "boxes": np.random.random(size=(2, 3, 4)).astype("float32"),
            "classes": np.random.random(size=(2, 3)).astype("float32"),
        }
        keypoints = np.random.random(size=(2, 5, 2)).astype("float32")
        segmentation_masks = np.random.random(size=(2, 8, 8, 1)).astype(
            "float32"
        )

        output = add_layer(
            {
                "images": images,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_masks,
            }
        )

        bounding_boxes_diff = (
            output["bounding_boxes"]["boxes"] - bounding_boxes["boxes"]
        )
        keypoints_diff = output["keypoints"] - keypoints
        segmentation_mask_diff = (
            output["segmentation_masks"] - segmentation_masks
        )
        self.assertNotAllClose(bounding_boxes_diff[0], bounding_boxes_diff[1])
        self.assertNotAllClose(keypoints_diff[0], keypoints_diff[1])
        self.assertNotAllClose(
            segmentation_mask_diff[0], segmentation_mask_diff[1]
        )

        @tf.function
        def in_tf_function(inputs):
            return add_layer(inputs)

        output = in_tf_function(
            {
                "images": images,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_masks,
            }
        )

        bounding_boxes_diff = (
            output["bounding_boxes"]["boxes"] - bounding_boxes["boxes"]
        )
        keypoints_diff = output["keypoints"] - keypoints
        segmentation_mask_diff = (
            output["segmentation_masks"] - segmentation_masks
        )
        self.assertNotAllClose(bounding_boxes_diff[0], bounding_boxes_diff[1])
        self.assertNotAllClose(keypoints_diff[0], keypoints_diff[1])
        self.assertNotAllClose(
            segmentation_mask_diff[0], segmentation_mask_diff[1]
        )

    def test_augment_all_data_in_tf_function(self):
        add_layer = VectorizedRandomAddLayer()
        images = np.random.random(size=(2, 8, 8, 3)).astype("float32")
        bounding_boxes = {
            "boxes": np.random.random(size=(2, 3, 4)).astype("float32"),
            "classes": np.random.random(size=(2, 3)).astype("float32"),
        }
        keypoints = np.random.random(size=(2, 5, 2)).astype("float32")
        segmentation_masks = np.random.random(size=(2, 8, 8, 1)).astype(
            "float32"
        )

        @tf.function
        def in_tf_function(inputs):
            return add_layer(inputs)

        output = in_tf_function(
            {
                "images": images,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_masks,
            }
        )

        bounding_boxes_diff = (
            output["bounding_boxes"]["boxes"] - bounding_boxes["boxes"]
        )
        keypoints_diff = output["keypoints"] - keypoints
        segmentation_mask_diff = (
            output["segmentation_masks"] - segmentation_masks
        )
        self.assertNotAllClose(bounding_boxes_diff[0], bounding_boxes_diff[1])
        self.assertNotAllClose(keypoints_diff[0], keypoints_diff[1])
        self.assertNotAllClose(
            segmentation_mask_diff[0], segmentation_mask_diff[1]
        )

    def test_augment_unbatched_all_data(self):
        add_layer = VectorizedRandomAddLayer(fixed_value=2.0)
        images = np.random.random(size=(8, 8, 3)).astype("float32")
        bounding_boxes = {
            "boxes": np.random.random(size=(3, 4)).astype("float32"),
            "classes": np.random.random(size=(3)).astype("float32"),
        }
        keypoints = np.random.random(size=(5, 2)).astype("float32")
        segmentation_masks = np.random.random(size=(8, 8, 1)).astype("float32")
        input = {
            "images": images,
            "bounding_boxes": bounding_boxes,
            "keypoints": keypoints,
            "segmentation_masks": segmentation_masks,
        }

        output = add_layer(input, training=True)
        expected_output = {
            "images": images + 2.0,
            "bounding_boxes": bounding_box.to_dense(
                {
                    "boxes": bounding_boxes["boxes"] + 2.0,
                    "classes": bounding_boxes["classes"] + 2.0,
                }
            ),
            "keypoints": keypoints + 2.0,
            "segmentation_masks": segmentation_masks + 2.0,
        }

        self.assertAllClose(output["images"], expected_output["images"])
        self.assertAllClose(output["keypoints"], expected_output["keypoints"])
        self.assertAllClose(
            output["bounding_boxes"]["boxes"],
            expected_output["bounding_boxes"]["boxes"],
        )
        self.assertAllClose(
            output["bounding_boxes"]["classes"],
            expected_output["bounding_boxes"]["classes"],
        )
        self.assertAllClose(
            output["segmentation_masks"], expected_output["segmentation_masks"]
        )

    def test_augment_all_data_for_assertion(self):
        images = np.random.random(size=(2, 8, 8, 3)).astype("float32")
        labels = np.squeeze(np.eye(10)[np.array([0, 1]).reshape(-1)])
        bounding_boxes = {
            "boxes": np.random.random(size=(2, 3, 4)).astype("float32"),
            "classes": np.random.random(size=(2, 3)).astype("float32"),
        }
        keypoints = np.random.random(size=(2, 5, 2)).astype("float32")
        segmentation_masks = np.random.random(size=(2, 8, 8, 1)).astype(
            "float32"
        )
        custom_annotations = np.random.random(size=(2, 1)).astype("float32")
        assertion_layer = VectorizedAssertionLayer()

        _ = assertion_layer(
            {
                "images": images,
                "labels": labels,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_masks,
                "custom_annotations": custom_annotations,
            }
        )

        # assertion is at VectorizedAssertionLayer's methods

    def test_augment_all_data_with_ragged_images_for_assertion(self):
        images = tf.ragged.stack(
            [
                np.random.random(size=(8, 8, 3)).astype("float32"),
                np.random.random(size=(16, 8, 3)).astype("float32"),
            ]
        )
        labels = np.squeeze(np.eye(10)[np.array([0, 1]).reshape(-1)])
        bounding_boxes = {
            "boxes": np.random.random(size=(2, 3, 4)).astype("float32"),
            "classes": np.random.random(size=(2, 3)).astype("float32"),
        }
        keypoints = np.random.random(size=(2, 5, 2)).astype("float32")
        segmentation_masks = np.random.random(size=(2, 8, 8, 1)).astype(
            "float32"
        )
        custom_annotations = np.random.random(size=(2, 1)).astype("float32")
        assertion_layer = VectorizedAssertionLayer()

        _ = assertion_layer(
            {
                "images": images,
                "labels": labels,
                "bounding_boxes": bounding_boxes,
                "keypoints": keypoints,
                "segmentation_masks": segmentation_masks,
                "custom_annotations": custom_annotations,
            }
        )

        # assertion is at VectorizedAssertionLayer's methods

    def test_converts_ragged_to_dense_images(self):
        images = tf.ragged.stack(
            [
                np.random.random(size=(8, 8, 3)).astype("float32"),
                np.random.random(size=(16, 8, 3)).astype("float32"),
            ]
        )
        add_layer = VectorizedRandomAddLayer(fixed_value=0.5)
        add_layer.force_output_dense_images = True

        result = add_layer(images)

        self.assertTrue(isinstance(result, tf.Tensor))

    def test_converts_ragged_to_dense_segmention_masks(self):
        images = tf.ragged.stack(
            [
                np.random.random(size=(8, 8, 3)).astype("float32"),
                np.random.random(size=(16, 8, 3)).astype("float32"),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                np.random.randint(0, 10, size=(8, 8, 1)).astype("float32"),
                np.random.randint(0, 10, size=(16, 8, 1)).astype("float32"),
            ]
        )
        add_layer = VectorizedRandomAddLayer(fixed_value=0.5)
        add_layer.force_output_dense_images = True

        result = add_layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))



import tensorflow as tf

from keras_aug import layers


class InvertTest(tf.test.TestCase):
    regular_args = {
        "value_range": (0, 255),
    }

    def test_invert_output(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        expected_output = 255.0 - image

        layer = layers.Invert(**self.regular_args)
        output = layer(image)

        self.assertAllClose(output, expected_output)



import tensorflow as tf

from keras_aug.datapoints import bounding_box as bounding_box_utils


class BoundingBoxUtilsTest(tf.test.TestCase):
    def test_sanitize_bounding_boxes_intact(self):
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[[10, 10, 20, 20], [10, 10, 30, 30]]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([[1, 2]], dtype=tf.float32),
        }
        expected_classes = tf.convert_to_tensor([[1, 2]], dtype=tf.float32)

        result_bounding_boxes = bounding_box_utils.sanitize_bounding_boxes(
            bounding_boxes
        )

        self.assertAllEqual(result_bounding_boxes["classes"], expected_classes)

    def test_sanitize_bounding_boxes_by_min_size(self):
        images = tf.ones((1, 50, 50, 3))
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[[10, 10, 20, 20], [10, 10, 30, 30]]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([[1, 2]], dtype=tf.float32),
        }
        expected_classes = tf.convert_to_tensor([[-1, 2]], dtype=tf.float32)

        result_bounding_boxes = bounding_box_utils.sanitize_bounding_boxes(
            bounding_boxes,
            min_size=25,
            bounding_box_format="xywh",
            images=images,
        )

        self.assertAllEqual(result_bounding_boxes["classes"], expected_classes)

    def test_sanitize_bounding_boxes_by_min_area_ratio(self):
        images = tf.ones((1, 50, 50, 3))
        ref_images = tf.ones((1, 50, 50, 3))
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[[10, 10, 20, 20], [10, 10, 30, 30]]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([[1, 2]], dtype=tf.float32),
        }
        ref_bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[[10, 10, 25, 25], [10, 10, 100, 100]]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([[1, 2]], dtype=tf.float32),
        }
        expected_classes = tf.convert_to_tensor([[1, -1]], dtype=tf.float32)

        result_bounding_boxes = bounding_box_utils.sanitize_bounding_boxes(
            bounding_boxes,
            min_area_ratio=0.1,
            bounding_box_format="xywh",
            reference_bounding_boxes=ref_bounding_boxes,
            images=images,
            reference_images=ref_images,
        )

        self.assertAllEqual(result_bounding_boxes["classes"], expected_classes)

    def test_sanitize_bounding_boxes_by_max_aspect_ratio(self):
        images = tf.ones((1, 50, 50, 3))
        bounding_boxes = {
            "boxes": tf.convert_to_tensor(
                [[[10, 10, 20, 20], [10, 10, 30, 3001]]], dtype=tf.float32
            ),
            "classes": tf.convert_to_tensor([[1, 2]], dtype=tf.float32),
        }
        expected_classes = tf.convert_to_tensor([[1, -1]], dtype=tf.float32)

        result_bounding_boxes = bounding_box_utils.sanitize_bounding_boxes(
            bounding_boxes,
            max_aspect_ratio=100.0,
            bounding_box_format="xywh",
            images=images,
        )

        self.assertAllEqual(result_bounding_boxes["classes"], expected_classes)



from keras_aug.datapoints.image.fill import corners_to_mask
from keras_aug.datapoints.image.fill import fill_rectangle
from keras_aug.datapoints.image.get_matrix import get_rotation_matrix
from keras_aug.datapoints.image.get_matrix import get_shear_matrix
from keras_aug.datapoints.image.get_matrix import get_translation_matrix
from keras_aug.datapoints.image.get_matrix import get_zoom_matrix
from keras_aug.datapoints.image.ops import blend
from keras_aug.datapoints.image.ops import rgb_to_grayscale
from keras_aug.datapoints.image.ops import transform
from keras_aug.datapoints.image.ops import transform_value_range



import tensorflow as tf
from tensorflow import keras

from keras_aug import core
from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomCropAndResize(VectorizedBaseRandomLayer):
    """Randomly crops a part of an image and resizes it to provided size.

    This implementation takes an intuitive approach, where we crop the images to
    a random height and width, and then resize them. To do this, we first sample
    a random value for area using `crop_area_factor` and a value for aspect
    ratio using `aspect_ratio_factor`. Further we get the new height and width
    by dividing and multiplying the old height and width by the random area
    respectively. We then sample offsets for height and width and clip them such
    that the cropped area does not exceed image boundaries. Finally, we do the
    actual cropping operation and resize the image to (`height`, `width`).

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        crop_area_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range of the area of the cropped part to that of original image.
            For self-supervised pretraining a common value for this parameter is
            ``(0.08, 1.0)``. For fine-tuning and classification a common value
            is ``(0.8, 1.0)``.
        aspect_ratio_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            ratio of width to height of the cropped image. When represented as
            a single float, the factor will be picked between
            ``[1.0 - factor, 1.0]``. For most tasks, this should be
            ``(3/4, 4/3)``.
        interpolation (str, optional): The interpolation mode. Supported values:
            ``"nearest", "bilinear"``. Defaults to `"bilinear"`.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        crop_area_factor,
        aspect_ratio_factor,
        interpolation="bilinear",
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self._check_arguments(
            height, width, crop_area_factor, aspect_ratio_factor
        )
        self.height = height
        self.width = width
        self.aspect_ratio_factor = augmentation_utils.parse_factor(
            aspect_ratio_factor,
            min_value=0.0,
            max_value=None,
            center_value=1.0,
            seed=seed,
        )
        if isinstance(crop_area_factor, float):
            lower = 1.0 - crop_area_factor
            upper = 1.0
            crop_area_factor = (lower, upper)
        self.crop_area_factor = augmentation_utils.parse_factor(
            crop_area_factor,
            min_value=0.0,
            max_value=1.0,
            seed=seed,
        )
        self.interpolation = interpolation
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_output_dense_images = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        crop_area_factor = self.crop_area_factor(shape=(batch_size, 1))
        aspect_ratio = self.aspect_ratio_factor(shape=(batch_size, 1))

        new_height = tf.clip_by_value(
            tf.sqrt(crop_area_factor / aspect_ratio), 0.0, 1.0
        )  # to avoid unwanted/unintuitive effects
        new_width = tf.clip_by_value(
            tf.sqrt(crop_area_factor * aspect_ratio), 0.0, 1.0
        )

        height_offset = self._random_generator.random_uniform(
            shape=(batch_size, 1),
            minval=tf.minimum(0.0, 1.0 - new_height),
            maxval=tf.maximum(0.0, 1.0 - new_height),
            dtype=tf.float32,
        )

        width_offset = self._random_generator.random_uniform(
            shape=(batch_size, 1),
            minval=tf.minimum(0.0, 1.0 - new_width),
            maxval=tf.maximum(0.0, 1.0 - new_width),
            dtype=tf.float32,
        )

        y1s = height_offset
        y2s = height_offset + new_height
        x1s = width_offset
        x2s = width_offset + new_width

        boxes = tf.concat([y1s, x1s, y2s, x2s], axis=-1)
        return boxes

    def compute_ragged_image_signature(self, images):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        batch_size = tf.shape(images)[0]
        boxes = transformations
        indices = tf.range(batch_size)

        # tf.image.crop_and_resize not support bfloat16
        if images.dtype == tf.bfloat16:
            images = tf.cast(images, dtype=tf.float32)
        images = tf.image.crop_and_resize(
            images,  # image shape: [B, H, W, C]
            boxes,  # boxes: (B, 4) in this case; represents area
            # to be cropped from the original image
            indices,  # box_indices: maps boxes to images along batch axis
            [self.height, self.width],  # output size
            method=self.interpolation,
        )
        images = tf.ensure_shape(
            images, shape=(None, self.height, self.width, None)
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomCropAndResize()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomCropAndResize(bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="rel_xyxy",
            images=raw_images,
        )

        t_y1s, t_x1s, t_y2s, t_x2s = tf.split(transformations, 4, axis=-1)
        # broadcast
        t_y1s, t_x1s = t_y1s[..., tf.newaxis], t_x1s[..., tf.newaxis]
        t_y2s, t_x2s = t_y2s[..., tf.newaxis], t_x2s[..., tf.newaxis]

        t_dxs = t_x2s - t_x1s
        t_dys = t_y2s - t_y1s
        x1s, y1s, x2s, y2s = tf.split(bounding_boxes["boxes"], 4, axis=-1)
        output = tf.concat(
            [
                (x1s - t_x1s) / t_dxs,
                (y1s - t_y1s) / t_dys,
                (x2s - t_x1s) / t_dxs,
                (y2s - t_y1s) / t_dys,
            ],
            axis=-1,
        )
        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = output

        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="rel_xyxy",
            images=images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="rel_xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def compute_ragged_segmentation_mask_signature(self, segmentation_masks):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, segmentation_masks.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        batch_size = tf.shape(segmentation_masks)[0]
        boxes = transformations
        indices = tf.range(batch_size)

        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if segmentation_masks.dtype == tf.bfloat16:
            segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        segmentation_masks = tf.image.crop_and_resize(
            segmentation_masks,  # image shape: [B, H, W, C]
            boxes,  # boxes: (B, 4) in this case; represents area
            # to be cropped from the original image
            indices,  # box_indices: maps boxes to images along batch axis
            [self.height, self.width],  # output size
            method="nearest",
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=(None, self.height, self.width, None)
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def _check_arguments(
        self, height, width, crop_area_factor, aspect_ratio_factor
    ):
        if not isinstance(height, int) or not isinstance(width, int):
            raise ValueError(
                "`height` and `width` must be integer. Received: "
                f"`height`={height} `width`={width} "
            )

        if (
            not isinstance(crop_area_factor, (tuple, list, core.FactorSampler))
            or isinstance(crop_area_factor, float)
            or isinstance(crop_area_factor, int)
        ):
            raise ValueError(
                "`crop_area_factor` must be tuple of two positive floats less "
                "than or equal to 1 or keras_aug.core.FactorSampler instance. "
                f"Received crop_area_factor={crop_area_factor}"
            )

        if (
            not isinstance(
                aspect_ratio_factor, (tuple, list, core.FactorSampler)
            )
            or isinstance(aspect_ratio_factor, float)
            or isinstance(aspect_ratio_factor, int)
        ):
            raise ValueError(
                "`aspect_ratio_factor` must be tuple of two positive floats or "
                "keras_aug.core.FactorSampler instance. Received "
                f"aspect_ratio_factor={aspect_ratio_factor}"
            )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "crop_area_factor": self.crop_area_factor,
                "aspect_ratio_factor": self.aspect_ratio_factor,
                "interpolation": self.interpolation,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class RandomChannelShiftTest(tf.test.TestCase, parameterized.TestCase):
    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((4, 8, 3)), tf.ones((4, 8, 3))],
                axis=0,
            ),
            dtype=tf.float32,
        )
        layer = layers.RandomChannelShift(
            factor=[0.1, 0.3], value_range=(0, 255)
        )

        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs[0] == 2.0))
        self.assertFalse(tf.math.reduce_any(xs[1] == 1.0))

    def test_5_channels(self):
        xs = tf.cast(tf.ones((4, 4, 5)), dtype=tf.float32)
        layer = layers.RandomChannelShift(
            factor=0.4, channels=5, value_range=(0, 255)
        )
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))

    def test_1_channel(self):
        xs = tf.cast(tf.ones((4, 4, 1)), dtype=tf.float32)
        layer = layers.RandomChannelShift(
            factor=0.4, channels=1, value_range=(0, 255)
        )
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))

    def test_in_single_image(self):
        xs = tf.cast(tf.ones((4, 4, 3)), dtype=tf.float32)
        layer = layers.RandomChannelShift(factor=0.4, value_range=(0, 255))
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))



from keras_aug.datapoints.bounding_box.converter import FROM_XYXY_CONVERTERS
from keras_aug.datapoints.bounding_box.converter import TO_XYXY_CONVERTERS
from keras_aug.datapoints.bounding_box.converter import convert_format
from keras_aug.datapoints.bounding_box.iou import compute_iou
from keras_aug.datapoints.bounding_box.to_dense import to_dense
from keras_aug.datapoints.bounding_box.to_ragged import to_ragged
from keras_aug.datapoints.bounding_box.utils import as_relative
from keras_aug.datapoints.bounding_box.utils import clip_to_image
from keras_aug.datapoints.bounding_box.utils import is_relative
from keras_aug.datapoints.bounding_box.utils import sanitize_bounding_boxes
from keras_aug.datapoints.bounding_box.validate_format import validate_format



import numpy as np
import tensorflow as tf

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class CenterCropTest(tf.test.TestCase):
    height, width = 224, 224
    regular_args = {
        "height": 112,
        "width": 112,
        "bounding_box_format": "rel_xyxy",
    }
    no_aug_args = {
        "height": 224,
        "width": 224,
        "bounding_box_format": "rel_xyxy",
    }

    def test_no_adjustment(self):
        image_shape = (4, self.height, self.width, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = layers.CenterCrop(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, self.height, self.width, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.CenterCrop(**self.no_aug_args)
        output = layer(image)
        self.assertAllClose(image, output, rtol=1e-5, atol=1e-5)

        layer = layers.CenterCrop(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_augments_image(self):
        input_image_shape = (4, self.height, self.width, 3)
        image = tf.random.uniform(shape=input_image_shape)
        layer = layers.CenterCrop(**self.regular_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertNotAllClose(output, input_image_resized)

    def test_grayscale(self):
        input_image_shape = (4, self.height, self.width, 1)
        image = tf.random.uniform(shape=input_image_shape)

        layer = layers.CenterCrop(**self.regular_args, seed=2023)
        input_image_resized = tf.image.resize(image, [self.height, self.width])

        output = layer(image)

        self.assertAllEqual(
            output.shape,
            (4, self.regular_args["height"], self.regular_args["width"], 1),
        )
        self.assertNotAllClose(output, input_image_resized)

    def test_augment_sparse_segmentation_mask(self):
        num_classes = 8
        input_image_shape = (1, self.height, self.width, 3)
        mask_shape = (1, self.height, self.width, 1)
        image = tf.random.uniform(shape=input_image_shape, seed=2023)
        mask = np.random.randint(2, size=mask_shape) * (num_classes - 1)
        inputs = {"images": image, "segmentation_masks": mask}

        # Crop-only to exactly 1/2 of the size
        args = self.no_aug_args.copy()
        args.update({"height": 112, "width": 112})
        layer = layers.CenterCrop(**args)
        input_mask_resized = tf.image.crop_and_resize(
            mask, [[0, 0, 1, 1]], [0], (112, 112), "nearest"
        )

        output = layer(inputs)

        self.assertNotAllClose(output["segmentation_masks"], input_mask_resized)

        # Crop to an arbitrary size and make sure we don't do bad interpolation
        args = self.no_aug_args.copy()
        args.update({"height": 123, "width": 123})
        layer = layers.CenterCrop(**args)

        output = layer(inputs)

        self.assertAllInSet(output["segmentation_masks"], [0, num_classes - 1])

    def test_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]]),
            "classes": tf.convert_to_tensor([0]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.no_aug_args.copy()
        args.update({"height": 10, "width": 10})
        layer = layers.CenterCrop(**args)
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }
        input = {"images": [image, image], "bounding_boxes": boxes}
        args = self.no_aug_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.CenterCrop(**args)
        expected_output = {
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
        }

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        input = {"images": image, "bounding_boxes": boxes}
        args = self.no_aug_args.copy()
        args.update({"height": 18, "width": 18})
        layer = layers.CenterCrop(**args)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant([[0, 0], [0]]),
        }
        expected_output = bounding_box.to_dense(expected_output)

        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )

        self.assertAllClose(
            expected_output["boxes"], output["bounding_boxes"]["boxes"]
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        args = self.no_aug_args.copy()
        args.update({"height": 4, "width": 4})
        layer = layers.CenterCrop(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (4, 4))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        args = self.no_aug_args.copy()
        args.update({"height": 4, "width": 4})
        layer = layers.CenterCrop(**args)

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(result["segmentation_masks"].shape[1:3], (4, 4))
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_input_with_graph_mode(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        args = self.regular_args.copy()
        args.update({"height": 8, "width": 8})
        layer = layers.CenterCrop(**args)

        @tf.function
        def fn(inputs):
            outputs = layer(inputs)
            image_shape = outputs["images"].shape
            segmentation_mask_shape = outputs["segmentation_masks"].shape
            assert image_shape == (2, 8, 8, 3)
            assert segmentation_mask_shape == (2, 8, 8, 1)

        fn({"images": images, "segmentation_masks": segmentation_masks})



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers


class RandAugmentTest(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(
        ("0", 0),
        ("5", 5),
        ("10", 10),
        ("20", 20),
        ("30", 30),
    )
    def test_runs_with_magnitude(self, magnitude):
        rand_augment = layers.RandAugment(
            value_range=(0, 255),
            magnitude=magnitude,
            seed=2023,
        )
        xs = tf.ones((2, 4, 4, 3))

        ys = rand_augment(xs)

        self.assertEqual(ys.shape, (2, 4, 4, 3))

    def test_runs_with_no_geometry(self):
        rand_augment = layers.RandAugment(
            value_range=(0, 255),
            magnitude=10,
            use_geometry=False,
            seed=2023,
        )
        xs = tf.ones((2, 4, 4, 3))

        ys = rand_augment(xs)

        self.assertEqual(ys.shape, (2, 4, 4, 3))

    def test_runs_with_exclude_ops(self):
        rand_augment = layers.RandAugment(
            value_range=(0, 255),
            magnitude=10,
            exclude_ops=["sharpness", "shear_y"],
            seed=2023,
        )
        xs = tf.ones((2, 4, 4, 3))

        ys = rand_augment(xs)

        self.assertEqual(ys.shape, (2, 4, 4, 3))



import inspect

import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.core import ConstantFactorSampler
from keras_aug.core import UniformFactorSampler
from keras_aug.layers import augmentation
from keras_aug.layers import preprocessing

GENERAL_TESTS = [
    (
        "AugMix",
        layers.AugMix,
        {"value_range": (0, 255)},
    ),
    (
        "RandAugment",
        layers.RandAugment,
        {"value_range": (0, 255)},
    ),
    (
        "TrivialAugmentWide",
        layers.TrivialAugmentWide,
        {"value_range": (0, 255)},
    ),
    (
        "RandomAffine",
        layers.RandomAffine,
        {
            "rotation_factor": 10,
            "translation_height_factor": 0.1,
            "translation_width_factor": 0.1,
            "zoom_height_factor": 0.1,
            "zoom_width_factor": 0.1,
            "shear_height_factor": 0.1,
            "shear_width_factor": 0.1,
        },
    ),
    (
        "RandomCrop",
        layers.RandomCrop,
        {"height": 2, "width": 2},
    ),
    (
        "RandomCropAndResize",
        layers.RandomCropAndResize,
        {
            "height": 2,
            "width": 2,
            "crop_area_factor": (0.8, 1.0),
            "aspect_ratio_factor": (3 / 4, 4 / 3),
        },
    ),
    (
        "RandomFlip",
        layers.RandomFlip,
        {"mode": "horizontal"},
    ),
    (
        "RandomResize",
        layers.RandomResize,
        {"heights": [2]},
    ),
    (
        "RandomRotate",
        layers.RandomRotate,
        {"factor": 10},
    ),
    (
        "RandomZoomAndCrop",
        layers.RandomZoomAndCrop,
        {"height": 2, "width": 2, "scale_factor": (0.8, 1.25)},
    ),
    (
        "CenterCrop",
        layers.CenterCrop,
        {"height": 2, "width": 2},
    ),
    (
        "PadIfNeeded",
        layers.PadIfNeeded,
        {"min_height": 2, "min_width": 2},
    ),
    (
        "ChannelShuffle",
        layers.ChannelShuffle,
        {"groups": 3},
    ),
    (
        "RandomBlur",
        layers.RandomBlur,
        {"factor": (3, 7)},
    ),
    (
        "RandomChannelShift",
        layers.RandomChannelShift,
        {"value_range": (0, 255), "factor": 0.1},
    ),
    (
        "RandomCLAHE",
        layers.RandomCLAHE,
        {"value_range": (0, 255)},
    ),
    (
        "RandomColorJitter",
        layers.RandomColorJitter,
        {
            "value_range": (0, 255),
            "brightness_factor": 0.1,
            "contrast_factor": 0.1,
            "saturation_factor": 0.1,
            "hue_factor": 0.1,
        },
    ),
    (
        "RandomGamma",
        layers.RandomGamma,
        {"value_range": (0, 255), "factor": 0.1},
    ),
    (
        "RandomGaussianBlur",
        layers.RandomGaussianBlur,
        {"kernel_size": 3, "factor": 2.0},
    ),
    (
        "RandomHSV",
        layers.RandomHSV,
        {
            "value_range": (0, 255),
            "hue_factor": 0.1,
            "saturation_factor": 0.1,
            "value_factor": 0.1,
        },
    ),
    (
        "RandomJpegQuality",
        layers.RandomJpegQuality,
        {
            "value_range": (0, 255),
            "factor": (75, 100),
        },
    ),
    (
        "RandomPosterize",
        layers.RandomPosterize,
        {"value_range": (0, 255), "factor": (5, 8)},
    ),
    (
        "RandomSharpness",
        layers.RandomSharpness,
        {"value_range": (0, 255), "factor": 0.1},
    ),
    (
        "RandomSolarize",
        layers.RandomSolarize,
        {
            "value_range": (0, 255),
            "threshold_factor": 10,
            "addition_factor": 10,
        },
    ),
    (
        "CutMix",
        layers.CutMix,
        {"alpha": 1.0},
    ),
    (
        "MixUp",
        layers.MixUp,
        {},
    ),
    (
        "Mosaic",
        layers.Mosaic,
        {
            "height": 2,
            "width": 2,
        },
    ),
    (
        "RandomChannelDropout",
        layers.RandomChannelDropout,
        {},
    ),
    (
        "RandomCutout",
        layers.RandomCutout,
        {"height_factor": 0.3, "width_factor": 0.3},
    ),
    (
        "RandomErase",
        layers.RandomErase,
        {"area_factor": (0.02, 0.4), "aspect_ratio_factor": (0.3, 1.0 / 0.3)},
    ),
    (
        "RandomGridMask",
        layers.RandomGridMask,
        {
            "size_factor": (0.5, 1.0),
            "ratio_factor": (0.6, 0.6),
            "rotation_factor": (-10, 10),
        },
    ),
    (
        "RandomApply",
        layers.RandomApply,
        {"layer": layers.RandomChannelDropout()},
    ),
    (
        "RandomChoice",
        layers.RandomChoice,
        {
            "layers": [
                layers.RandomChannelDropout(),
                layers.RandomChannelDropout(),
            ]
        },
    ),
    (
        "RepeatedAugment",
        layers.RepeatedAugment,
        {
            "layers": [
                layers.RandomColorJitter(
                    value_range=(0, 255), brightness_factor=(1.5, 1.5)
                ),
                layers.RandomColorJitter(
                    value_range=(0, 255), contrast_factor=(1.5, 1.5)
                ),
            ]
        },
    ),
    (
        "Resize",
        layers.Resize,
        {"height": 2, "width": 2},
    ),
    (
        "AutoContrast",
        layers.AutoContrast,
        {"value_range": (0, 255)},
    ),
    (
        "Equalize",
        layers.Equalize,
        {"value_range": (0, 255)},
    ),
    (
        "Grayscale",
        layers.Grayscale,
        {"output_channels": 3},
    ),
    (
        "Identity",
        layers.Identity,
        {},
    ),
    (
        "Invert",
        layers.Invert,
        {"value_range": (0, 255)},
    ),
    (
        "Normalize",
        layers.Normalize,
        {"value_range": (0, 255)},
    ),
    (
        "Rescale",
        layers.Rescale,
        {"scale": 1.0 / 255.0},
    ),
    (
        "SanitizeBoundingBox",
        layers.SanitizeBoundingBox,
        {"min_size": 10, "bounding_box_format": "xyxy"},
    ),
]


class ConfigTest(tf.test.TestCase, parameterized.TestCase):
    def test_all_2d_aug_layers_included(self):
        base_cls = layers.VectorizedBaseRandomLayer
        cls_spaces = [augmentation, preprocessing]
        all_2d_aug_layers = []
        for cls_space in cls_spaces:
            all_2d_aug_layers.extend(
                inspect.getmembers(cls_space, predicate=inspect.isclass)
            )
        all_2d_aug_layer_names = set(
            item[0]
            for item in all_2d_aug_layers
            if issubclass(item[1], base_cls)
        )

        general_names = set(item[0] for item in GENERAL_TESTS)
        all_test_names = general_names

        for name in all_2d_aug_layer_names:
            self.assertIn(name, all_test_names, msg=f"{name} not found")

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_config(self, layer_cls, args):
        layer = layer_cls(**args)

        config = layer.get_config()

        for key in args.keys():
            if isinstance(config[key], UniformFactorSampler):
                self.assertTrue(isinstance(config[key], UniformFactorSampler))
            elif isinstance(config[key], ConstantFactorSampler):
                self.assertTrue(isinstance(config[key], ConstantFactorSampler))
            else:
                self.assertEqual(config[key], args[key])

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_config_with_custom_name(self, layer_cls, args):
        layer = layer_cls(**args, name="image_preproc")
        config = layer.get_config()

        layer_1 = layer_cls.from_config(config)

        self.assertEqual(layer_1.name, layer.name)



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import core
from keras_aug.core.keras_random_generator import KerasRandomGenerator
from keras_aug.utils import augmentation as augmentation_utils


class AugmentationUtilsTest(tf.test.TestCase, parameterized.TestCase):
    def test_get_padding_position_invalid(self):
        with self.assertRaises(NotImplementedError):
            augmentation_utils.get_padding_position("hello")

    @parameterized.parameters(
        [
            "center",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
            "random",
        ]
    )
    def test_get_padding_position(self, position):
        _ = augmentation_utils.get_padding_position(position)

    @parameterized.named_parameters(
        *[
            ("center", ([1], [2], [3], [4]), ([1], [2], [3], [4]), "center"),
            (
                "top_left",
                ([1], [2], [3], [4]),
                ([0], [3], [0], [7]),
                "top_left",
            ),
            (
                "top_right",
                ([1], [2], [3], [4]),
                ([0], [3], [7], [0]),
                "top_right",
            ),
            (
                "bottom_left",
                ([1], [2], [3], [4]),
                ([3], [0], [0], [7]),
                "bottom_left",
            ),
            (
                "bottom_right",
                ([1], [2], [3], [4]),
                ([3], [0], [7], [0]),
                "bottom_right",
            ),
            ("random", ([1], [2], [3], [4]), None, "random"),
        ]
    )
    def test_get_position_params(self, values, expected_values, position_name):
        rng = KerasRandomGenerator(2023)
        tops, bottoms, lefts, rights = values
        position = augmentation_utils.get_padding_position(position_name)

        tops, bottoms, lefts, rights = augmentation_utils.get_position_params(
            tops, bottoms, lefts, rights, position, rng
        )
        if position_name == "random":
            return

        self.assertAllEqual((tops, bottoms, lefts, rights), expected_values)

    def test_is_factor_working_false(self):
        # int
        self.assertEqual(
            augmentation_utils.is_factor_working(0, not_working_value=0), False
        )
        # tuple
        self.assertEqual(
            augmentation_utils.is_factor_working((0, 0), not_working_value=0),
            False,
        )
        # core.ConstantFactorSampler
        self.assertEqual(
            augmentation_utils.is_factor_working(
                core.ConstantFactorSampler(0), not_working_value=0
            ),
            False,
        )
        # core.UniformFactorSampler
        self.assertEqual(
            augmentation_utils.is_factor_working(
                core.UniformFactorSampler(0, 0), not_working_value=0
            ),
            False,
        )

    def test_is_factor_working_true(self):
        # int
        self.assertEqual(
            augmentation_utils.is_factor_working(1, not_working_value=0), True
        )
        # tuple
        self.assertEqual(
            augmentation_utils.is_factor_working((1, 1), not_working_value=0),
            True,
        )
        # core.ConstantFactorSampler
        self.assertEqual(
            augmentation_utils.is_factor_working(
                core.ConstantFactorSampler(1), not_working_value=0
            ),
            True,
        )
        # core.UniformFactorSampler
        self.assertEqual(
            augmentation_utils.is_factor_working(
                core.UniformFactorSampler(0, 1), not_working_value=0
            ),
            True,
        )

    def test_expand_dict_dims(self):
        key = "factor"
        shape = [3, 3]
        transformation = {key: tf.random.uniform(shape=shape)}

        result = augmentation_utils.expand_dict_dims(transformation, axis=0)

        self.assertEqual(result[key].shape, [1] + [3, 3])

    def test_get_images_shape_dense(self):
        height = 5
        width = 6
        images = tf.random.uniform(shape=(2, 5, 6, 3))

        heights, widths = augmentation_utils.get_images_shape(images)

        self.assertAllEqual(heights, [[height], [height]])
        self.assertAllEqual(widths, [[width], [width]])

    def test_get_images_shape_ragged(self):
        images = tf.ragged.stack(
            [
                tf.ones((5, 5, 3)),
                tf.ones((8, 8, 3)),
            ]
        )

        heights, widths = augmentation_utils.get_images_shape(images)

        self.assertAllEqual(heights, [[5], [8]])
        self.assertAllEqual(widths, [[5], [8]])

    def test_cast_to_float16(self):
        inputs = {
            augmentation_utils.IMAGES: tf.random.uniform((2, 4, 4, 3)),
            augmentation_utils.LABELS: tf.random.uniform((2, 1)),
            augmentation_utils.BOUNDING_BOXES: {
                "boxes": tf.zeros((2, 2, 4)),
                "classes": tf.zeros((2, 2)),
            },
            augmentation_utils.SEGMENTATION_MASKS: tf.random.uniform(
                (2, 4, 4, 10)
            ),
            augmentation_utils.KEYPOINTS: tf.random.uniform((2, 4, 17)),
        }

        result = augmentation_utils.cast_to(inputs, dtype=tf.float16)

        self.assertDTypeEqual(result[augmentation_utils.IMAGES], tf.float16)
        self.assertDTypeEqual(result[augmentation_utils.LABELS], tf.float16)
        self.assertDTypeEqual(
            result[augmentation_utils.BOUNDING_BOXES]["boxes"], tf.float16
        )
        self.assertDTypeEqual(
            result[augmentation_utils.BOUNDING_BOXES]["classes"], tf.float16
        )
        self.assertDTypeEqual(result[augmentation_utils.KEYPOINTS], tf.float16)

    def test_cast_to_int8(self):
        inputs = {
            augmentation_utils.IMAGES: tf.random.uniform((2, 4, 4, 3)),
            augmentation_utils.LABELS: tf.random.uniform((2, 1)),
            augmentation_utils.BOUNDING_BOXES: {
                "boxes": tf.zeros((2, 2, 4)),
                "classes": tf.zeros((2, 2)),
            },
            augmentation_utils.SEGMENTATION_MASKS: tf.random.uniform(
                (2, 4, 4, 10)
            ),
            augmentation_utils.KEYPOINTS: tf.random.uniform((2, 4, 17)),
        }

        result = augmentation_utils.cast_to(inputs, dtype=tf.int8)

        self.assertDTypeEqual(result[augmentation_utils.IMAGES], tf.int8)
        self.assertDTypeEqual(result[augmentation_utils.LABELS], tf.int8)
        self.assertDTypeEqual(
            result[augmentation_utils.BOUNDING_BOXES]["boxes"], tf.int8
        )
        self.assertDTypeEqual(
            result[augmentation_utils.BOUNDING_BOXES]["classes"], tf.int8
        )
        self.assertDTypeEqual(result[augmentation_utils.KEYPOINTS], tf.int8)

    def test_compute_signature(self):
        inputs = {
            augmentation_utils.IMAGES: tf.random.uniform((2, 4, 4, 3)),
            augmentation_utils.LABELS: tf.random.uniform((2, 1)),
            augmentation_utils.BOUNDING_BOXES: {
                "boxes": tf.zeros((2, 2, 4)),
                "classes": tf.zeros((2, 2)),
            },
            augmentation_utils.SEGMENTATION_MASKS: tf.random.uniform(
                (2, 4, 4, 10)
            ),
            augmentation_utils.KEYPOINTS: tf.random.uniform((2, 4, 17)),
        }

        signatures = augmentation_utils.compute_signature(inputs, tf.float16)

        self.assertTrue(inputs.keys() == signatures.keys())
